# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for interacting with use flag configuration files

These are files containing lines of the form:
    category/package-name flag -otherflag

Atoms in these files should always be qualified, but may be as general
or as specific as the specification allows, optionally using operators to match
specific versions or ranges of versions.
"""

import os
import re
from logging import info
from typing import Iterable, Optional, Set

from portmod.functools import system_cache
from portmodlib.atom import Atom, atom_sat
from portmodlib.l10n import l10n
from portmodlib.parsers.list import read_list


def get_flags(file: str, atom: Optional[Atom] = None) -> Set[str]:
    """
    Reads flags from a given file.

    args:
        file: Path of the file to be parsed
        atom: If non-None, file is assumed to be newline delimited
            with an atom followed by a list of flags on each line.
            If None, file is assumed to be a newline delimited list of flags
    returns:
        The set of flags contained in the file and matching the atom
    """
    flags: Set[str] = set()
    if os.path.exists(file):
        for line in read_list(file):
            # Remove comments
            line = re.sub("#.*", "", line)
            if not line:
                continue

            if atom:
                elem = line.split()
                line_atom = Atom(elem[0])
                if atom_sat(atom, line_atom):
                    flags = collapse_flags(flags, elem[1:])
            else:
                flags = collapse_flags(flags, [line])
        return flags
    return set()


def collapse_flags(old: Iterable[str], new: Iterable[str]) -> Set[str]:
    """
    Collases an ordered list of flags into an unordered set of flags

    The resulting set will contain only the last variant of a flag (enabled or
    disabled) present in the lists.
    """
    newset = set(old)
    for flag in new:
        if not flag.startswith("-"):
            newset.discard(f"-{flag}")
        elif flag.startswith("-"):
            newset.discard(flag.lstrip("-"))
        newset.add(flag)
    return newset


def add_flag(file: str, atom: Atom, flag: str):
    """
    Adds flag to the flag file

    args:
        file: Path of the file to be modified
        atom: Atom that the flag applies to. This function will only modify an existing
              atom/flag list if the atom exactly matches this atom, otherwise a new
              list will be added
        flag: Flag to be added, including the `-` prefix if the flag is to be
              disabled explicitly
    """
    if os.path.exists(file):
        flagfile = __read_flags(file)
    else:
        flagfile = []

    found = False
    for (index, line) in enumerate(flagfile):
        line = re.sub("#.*", "", line).strip()
        if not line:
            continue

        tokens = line.split()
        if tokens[0] == atom:
            if flag not in tokens:
                info(l10n("flag-add", flag=flag, atom=atom, file=file))
                flagfile[index] = "{} {}".format(line, flag)
            found = True
    if not found:
        info(l10n("flag-add", flag=flag, atom=atom, file=file))
        flagfile.append("{} {}".format(atom, flag))

    __write_flags(file, flagfile)


def remove_flag(file: str, atom: Atom, flag: str):
    """
    Removes flag from the flag file

    args:
        file: Path of the file to modify
        atom: Atom to match against. The flag will be disabled for all atoms in the
              file which satisfy this atom
        flag: Flag to remove from the file. The flag must match exactly, this will not
              also remove the disabled variant of a flag.
    """
    flagfile = __read_flags(file)

    for (index, line) in enumerate(flagfile):
        line = re.sub("#.*", "", line).strip()
        if not line:
            continue

        tokens = line.split()
        if atom_sat(Atom(tokens[0]), atom) and flag in tokens:
            info(l10n("flag-remove", flag=flag, atom=atom, file=file))
            tokens = list(filter(lambda a: a != flag, tokens))

            if len(tokens) > 1:
                flagfile[index] = " ".join(tokens)
            else:
                del flagfile[index]

    __write_flags(file, flagfile)


@system_cache
def __read_flags(file):
    if os.path.exists(file):
        with open(file, mode="r") as flagfile:
            return flagfile.read().splitlines()
    return []


def __write_flags(file, new_flagfile):
    with open(file, mode="w") as flagfile:
        for line in new_flagfile:
            print(line, file=flagfile)
    __read_flags.cache_clear()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import csv
import filecmp
import os
import shutil
from difflib import unified_diff
from logging import info
from pathlib import Path
from string import Template
from tempfile import NamedTemporaryFile
from typing import Generator, Iterable, List, Optional, Tuple

from portmodlib.execute import execute
from portmodlib.fs import match
from portmodlib.l10n import l10n
from portmodlib.parsers.list import add_list, read_list

from .config import get_config_value, variable_data_dir
from .globals import env
from .prompt import prompt_options


def get_protected_path(file: str) -> str:
    new_file = file
    i = 1
    while os.path.exists(new_file):
        name, ext = os.path.splitext(new_file)
        if ext.startswith(".__cfg_protect"):
            new_file = name + f".__cfg_protect_{i}__"
            i += 1
        else:
            new_file = file + ".__cfg_protect__"
    return new_file


def is_protected(file: str) -> bool:
    protected = get_config_value("CFG_PROTECT")
    if isinstance(protected, list):
        for pattern in protected:
            if match(Path(file).absolute().relative_to(env.prefix().ROOT), pattern):
                return True
    elif isinstance(protected, str):
        if match(Path(file).absolute().relative_to(env.prefix().ROOT), protected):
            return True
    return False


def get_mergetool() -> Optional[str]:
    """Returns user-configured mergetool"""
    result = get_config_value("MERGE_TOOL")
    if result:
        return str(result)
    return None


def merge_files(source: str, dest: str):
    """Invokes user-configured mergetool"""
    mergetool = get_mergetool()
    assert mergetool
    execute(
        Template(mergetool).substitute(
            orig=f'"{source}"', new=f'"{dest}"', merged=f'"{dest}"'
        )
    )


def get_redirections() -> Generator[Tuple[str, str], None, None]:
    """
    Iterates over all previously made file redirections and returns the (non-empty)
    results
    """
    if os.path.exists(os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv")):
        with open(
            os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv"), "r"
        ) as file:
            reader = csv.reader(file)
            for row in reader:
                dst = row[0]
                src = row[1]

                if os.path.exists(src) and os.stat(src).st_size != 0:
                    yield src, dst


def remove_redirection(src: str, dst: str):
    path = os.path.join(env.prefix().CONFIG_PROTECT_DIR, "cfg_protect.csv")
    if os.path.exists(path):
        with open(path, "w") as output_file:
            writer = csv.writer(output_file)
            with open(path, "r") as file:
                reader = csv.reader(file)
                for row in reader:
                    row_dst = row[0]
                    row_src = row[1]
                    if row_dst != dst and row_src != src:
                        writer.writerow(row)


def get_changes(src: str, dst: str) -> List[str]:
    if os.path.islink(src):
        src_lines = [l10n("symlink-to", path=os.readlink(src)) + "\n"]
    else:
        try:
            with open(src, "r") as src_file:
                src_lines = src_file.readlines()
        except UnicodeDecodeError:
            src_lines = ["<" + l10n("binary-data") + ">\n"]
    dst_lines = []
    if os.path.lexists(dst):
        if os.path.islink(dst):
            dst_lines = [l10n("symlink-to", path=os.readlink(dst)) + "\n"]
        else:
            try:
                with open(dst, "r") as dst_file:
                    dst_lines = dst_file.readlines()
            except UnicodeDecodeError:
                dst_lines = ["<" + l10n("binary-data") + ">\n"]

    return list(unified_diff(dst_lines, src_lines, dst, src))


def handle_cfg_protect(files: Iterable[Tuple[str, str]]):
    """Prompts user to allow changes to files made by modules"""
    whitelist_file = os.path.join(variable_data_dir(), "module-data", "file-whitelist")
    blacklist_file = os.path.join(variable_data_dir(), "module-data", "file-blacklist")
    blacklist = set()
    whitelist = set()
    if os.path.exists(whitelist_file):
        whitelist = set(read_list(whitelist_file))
    if os.path.exists(blacklist_file):
        blacklist = set(read_list(blacklist_file))

    # Display file changes to user and prompt
    for src, dst in files:
        original_src = src
        if os.path.exists(dst) and filecmp.cmp(src, dst, shallow=False):
            os.remove(src)
            continue

        if dst in blacklist:
            info(l10n("skipped-blacklisted-file", file=dst))
            os.remove(src)
            continue

        while True:
            output = get_changes(src, dst)
            if dst in whitelist:
                # User won't be prompted, so we should still display output, but supress it
                # unless running verbosely
                info("".join(output))
            else:
                print("".join(output))

            print()

            if dst not in whitelist and not env.INTERACTIVE:
                info(l10n("skipped-update-noninteractive", file=dst))
                break

            response = None
            if dst not in whitelist:
                options = [
                    (l10n("yes-short"), l10n("apply-change")),
                    (l10n("always-short"), l10n("merge-apply-always")),
                    (l10n("skip-change.short"), l10n("skip-change")),
                    (l10n("no-short"), l10n("merge-do-not-apply-change")),
                    (l10n("never-short"), l10n("merge-apply-never")),
                ]
                mergetool = get_mergetool()
                if mergetool:
                    options.append(
                        (
                            l10n("mergetool.short"),
                            l10n("mergetool", mergetool=f'"{mergetool}"'),
                        )
                    )
                else:
                    print(l10n("mergetool-info", var="MERGE_TOOL"))
                    print()

                try:
                    response = prompt_options(l10n("apply-above-change-qn"), options)
                except EOFError:
                    response = l10n("no.short")

            if response == l10n("skip-change.short"):
                break

            if dst in whitelist or response in (
                l10n("yes-short"),
                l10n("always-short"),
            ):
                os.makedirs(os.path.dirname(dst), exist_ok=True)
                if os.path.exists(dst) or os.path.islink(dst):
                    os.remove(dst)
                shutil.move(src, dst)
                # Remove the original file if src now points to a merged file
                if original_src != src:
                    os.remove(original_src)

                if response == l10n("always-short"):
                    add_list(whitelist_file, dst)

                remove_redirection(src, dst)
                break

            if response == l10n("mergetool.short"):
                with open(dst, "r") as file:
                    contents = file.read()
                with NamedTemporaryFile(
                    "w",
                    delete=False,
                    prefix="EDITME.",
                    suffix="." + os.path.basename(dst),
                ) as tempfile:
                    tempfile.write(contents)
                    tempfile_path = tempfile.name
                # Merge against intermediate file
                merge_files(src, tempfile_path)
                # Don't break, as we must let the user accept the merged changes first
                src = tempfile_path

            if response in {l10n("no-short"), l10n("never-short")}:
                if response == l10n("never-short"):
                    add_list(blacklist_file, dst)

                os.remove(src)
                remove_redirection(src, dst)
                break



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import csv
import os

from .globals import config_protect_dir


def _add_redirection(source: str, dest: str):
    protect_dir = os.path.join(config_protect_dir())
    csv_path = os.path.join(protect_dir, "cfg_protect.csv")
    if os.path.exists(csv_path):
        with open(csv_path, "r") as file:
            reader = csv.reader(file)

            if [dest, source] not in reader:
                with open(csv_path, "a", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerow([dest, source])
    else:
        os.makedirs(os.path.dirname(csv_path), exist_ok=True)
        with open(csv_path, "w", newline="") as file:
            writer = csv.writer(file)
            writer.writerow([dest, source])


def create_file(path):
    """
    Returns the path to be used for redirection
    and adds it to the list of redirected paths
    """
    protect_dir = os.path.join(config_protect_dir())
    new_path = os.path.join(protect_dir, os.path.basename(path) + ".cfg_protect")
    if os.path.lexists(new_path):
        num = 1
        new_path = new_path + "." + str(num)

        while os.path.lexists(new_path):
            num += 1
            new_path, _ = os.path.splitext(new_path)
            new_path = new_path + "." + str(num)

    os.makedirs(protect_dir, exist_ok=True)
    _add_redirection(new_path, path)

    return new_path



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Type hints for Native Rust Extension"""
from typing import Any, Dict, List, Optional, Set, Union

# Ideally this file could eventually be generated automatically
# See https://github.com/PyO3/pyo3/issues/510

class Group:
    group: str

class Person:
    """
    An individual maintainer
    Either name or email is required.
    """

    # Maintainer's Name or Pseudonym
    name: Optional[str]
    # Maintainer's Email
    email: Optional[str]
    # Description. Can be used to describe the status of maintainership
    desc: Optional[str]

class GroupDeclaration:
    desc: str
    members: List[Person]

Maintainer = Union[Group, Person]
Maintainers = Union[Maintainer, List[Maintainer]]

class Upstream:
    # maintainers/authors of the original mod.
    maintainer: Optional[Maintainers]
    # URL where a changelog for the mod can be found. Must be version independent
    changelog: Optional[str]
    # URL where the location of the upstream documentation can be found.
    # The link must not point to any third party documentation and must be version independent
    doc: Optional[str]
    # A place where bugs can be reported in the form of an URL or an e-mail address prefixed with mailto:
    bugs_to: Optional[str]

class PackageMetadata:
    # Description of the package
    longdescription: Optional[str]
    # Maintainer, or list of maintainers for the package
    maintainer: Optional[Maintainers]
    # Use flags and their descriptions. Key is the flag name, value is the description
    use: Dict[str, str]
    # Tags for searching
    tags: Set[str]
    # Description of the package's upstream information.
    upstream: Optional[Upstream]

class CategoryMetadata:
    # Description of the category.
    longdescription: str
    # Maintainer, or list of maintainers for the category
    maintainer: Optional[Maintainers]

class News:
    # A short descriptive title
    title: str
    # Author's name and email address, in the form Real Name <email@address>
    author: str
    # Translator's name and email address, in the form Real Name <email@address>
    translator: Optional[str]
    # Date of posting, in yyyy-mm-dd format
    posted: str
    revision: str
    # Only supported format is 2.0
    news_item_format: str
    # Contents of the news article
    body: str
    # Required installed packages for the news to be displayed
    display_if_installed: Optional[str]
    # Required keywords for the news to be displayed
    display_if_keyword: Optional[str]
    # Required profiles for the news to be displayed
    display_if_profile: Optional[str]

def l10n_lookup(locale: str, msg_id: str, kwargs: Dict[str, Any]) -> str: ...
def get_masters(filename: str) -> List[str]: ...
def parse_yaml_dict(filename: str) -> Dict[str, str]: ...
def parse_yaml_dict_dict(filename: str) -> Dict[str, Dict[str, str]]: ...
def parse_news(filename: str) -> News: ...
def parse_groups(filename: str) -> Dict[str, GroupDeclaration]: ...
def parse_category_metadata(filename: str) -> CategoryMetadata: ...
def parse_package_metadata(filename: str) -> PackageMetadata: ...
def _get_hash(filename: str, algs: List[str], buffer_size: int) -> List[str]: ...

class PackageIndexData:
    """Data class for package metadata returned by query results"""

    cpn: str
    prefix: str
    category: str
    package: str
    name: str
    desc: str
    homepage: Optional[str]
    other_homepages: List[str]
    license: Optional[str]
    longdescription: Optional[str]
    maintainers: List[str]
    upstream_maintainers: List[str]
    upstream_doc: Optional[str]
    upstream_bugs_to: Optional[str]
    upstream_changelog: Optional[str]
    tags: Set[str]

    def __init__(
        self, cpn: str, prefix: str, category: str, package: str, name: str, desc: str
    ): ...

def update_index(index_path: str, packages: List[PackageIndexData]): ...
def query(
    index_path: str, prefix: str, query: str, limit: int
) -> List[PackageIndexData]: ...

class _Directories:
    cache_dir: str
    download_dir: Optional[str]
    config_dir: str
    data_dir: str

directories: _Directories



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import argparse
import logging
import os
import sys
import traceback
from logging import error
from typing import Optional, cast

from portmod.config import config_to_string, get_config
from portmod.config.sets import get_system
from portmod.download import RemoteHashError, SourceUnfetchable
from portmod.globals import env, get_version, refresh_env
from portmod.loader import SandboxedError, load_installed_pkg
from portmod.prefix import get_prefixes
from portmod.transactions import get_usestrings
from portmodlib.atom import Atom
from portmodlib.l10n import l10n
from portmodlib.log import add_logging_arguments, init_logger

from .cfg_update import add_cfg_update_parser
from .destroy import add_destroy_parser
from .error import InputException
from .init import add_init_parser
from .merge import add_merge_parser
from .mirror import add_mirror_parser
from .query import add_query_parser
from .run import add_run_parser
from .search import add_search_parser
from .select import add_select_parser
from .sync import add_sync_parser
from .use import add_use_parser
from .validate import validate


def add_info_parser(subparsers, parents):
    def info(args):
        from git import Repo

        # Strip the user's home directory from paths and replace it with ~
        # This command is designed to produce information to be submitted with
        # bug reports, and the user's home directory is not a relevant piece of information.
        def strip_user(string: str) -> str:
            return string.replace(os.path.expanduser("~"), "~")

        print(f"Portmod {get_version()}")
        print(f"Python {sys.version}")
        print()

        print(l10n("info-repositories"))
        for repo in env.prefix().REPOS:
            gitrepo = Repo.init(repo.location)
            print("    ", "\n         ".join(strip_user(str(repo)).split(",")))
            print(
                "    ",
                l10n(
                    "info-repository-date", date=gitrepo.head.commit.committed_datetime
                ),
            )
            print("    ", l10n("info-repository-commit", commit=gitrepo.head.commit))
            print()

        generator = (
            load_installed_pkg(atom)
            for atom in get_system() | set(map(Atom, get_config()["INFO_PACKAGES"]))
            if load_installed_pkg(atom)
        )
        packages = set(filter(None, generator))
        length = 0
        if len(packages) > 0:
            length = max(len(x.CPN) for x in packages)
        for pkg in sorted(packages, key=lambda x: x.CPN):
            # FIXME: Display date package was installed for live packages
            usestrings = filter(
                None, get_usestrings(pkg, pkg.INSTALLED_USE, pkg.INSTALLED_USE, True)
            )
            padding = length - len(str(pkg.CPN))
            print(pkg.CPN + ":", " " * padding, pkg.PVR, *usestrings)
        print()

        # Print config values
        config = get_config()
        if args.verbose:
            config_string = config_to_string(config)
        else:
            config_string = config_to_string(
                {
                    entry: config[entry]
                    for entry in config
                    if entry in config["INFO_VARS"]
                }
            )

        print(strip_user(config_string))
        # Print hardcoded portmod paths
        print(strip_user(f"TMP_DIR = {env.TMP_DIR}"))
        print(strip_user(f"CACHE_DIR = {env.CACHE_DIR}"))
        print(strip_user(f"CONFIG_DIR = {env.prefix().CONFIG_DIR}"))
        print(strip_user(f"ROOT = {env.prefix().ROOT}"))
        sys.exit(0)

    parser = subparsers.add_parser("info", help=l10n("info-help"), parents=parents)
    parser.set_defaults(func=info)


def get_generic_parser():
    """
    Produces a generic argument parser for use with the documentation generation.

    Does not include specific prefixes, or architectures
    """
    import shutil
    from tempfile import gettempdir

    from portmod.globals import env, refresh_env
    from portmod.prefix import add_prefix

    OLD = env.__dict__
    TESTDIR = os.path.join(gettempdir(), "portmod.test")
    env.CONFIG_DIR = os.path.join(TESTDIR, "config")
    env.CACHE_DIR = os.path.join(TESTDIR, "cache")
    env.DATA_DIR = os.path.join(TESTDIR, "local")
    env.INTERACTIVE = False

    refresh_env()
    add_prefix("<example>", "test")
    refresh_env()

    try:
        parser = get_parser("<example>", generic=True)

    finally:
        shutil.rmtree(TESTDIR)
        env.__dict__ = OLD

    return parser


def get_parser(
    subcommand: Optional[str] = None,
    prefix_subcommand: Optional[str] = None,
    generic=False,
):
    common = argparse.ArgumentParser(add_help=False)
    add_logging_arguments(common)
    common.add_argument("--debug", help=argparse.SUPPRESS, action="store_true")
    common.add_argument(
        "--no-confirm", help=l10n("no-confirm-help"), action="store_true"
    )

    parser = argparse.ArgumentParser(description=l10n("description"), parents=[common])
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--version", help=l10n("version-help"), action="store_true")

    subparsers = parser.add_subparsers(dest="subcommand_name")

    if (
        subcommand is None or subcommand in get_prefixes() and prefix_subcommand is None
    ) and not generic:
        prefix_parsers_list = []

        def get_help_func(parser):
            func = parser.print_help
            return lambda args: func()

        for prefix in get_prefixes():
            prefix_parser = subparsers.add_parser(
                prefix,
                help=l10n("prefix-help", prefix=prefix),
                description=l10n("prefix-help", prefix=prefix),
                parents=[common],
            )
            prefix_parsers_list.append(
                prefix_parser.add_subparsers(dest="prefix_subcommand_name")
            )
            prefix_parser.set_defaults(func=get_help_func(prefix_parser))

        for subparser in prefix_parsers_list:
            subparser.add_parser("merge", help=l10n("merge-help"), add_help=False)
            subparser.add_parser("search", help=l10n("search-help"), add_help=False)
            subparser.add_parser("select", help=l10n("select-help"), add_help=False)
            subparser.add_parser("query", help=l10n("query-help"), add_help=False)
            subparser.add_parser("use", help=l10n("use-help"), add_help=False)
            subparser.add_parser(
                "conflict-ui", help=l10n("conflict-ui-help"), add_help=False
            )
            subparser.add_parser("info", help=l10n("info-help"), add_help=False)
            subparser.add_parser("validate", help=l10n("validate-help"), add_help=False)
            subparser.add_parser("destroy", help=l10n("destroy-help"), add_help=False)
            subparser.add_parser("run", help=l10n("run-help"), add_help=False)
            subparser.add_parser(
                "cfg-update", help=l10n("cfg-update-help"), add_help=False
            )
        subparsers.add_parser("mirror", help=l10n("mirror-help"), add_help=False)
    elif subcommand is not None and subcommand in get_prefixes() and not generic:
        prefix = subcommand

        prefix_parser = subparsers.add_parser(
            prefix, help=l10n("prefix-help", prefix=prefix), parents=[common]
        )
        prefix_parsers = prefix_parser.add_subparsers(dest="prefix_subcommand_name")
        prefix_parser.set_defaults(func=lambda args: prefix_parser.print_help())
        if prefix_subcommand == "merge":
            add_merge_parser(prefix_parsers, [common])
        elif prefix_subcommand == "search":
            add_search_parser(prefix_parsers, [common])
        elif prefix_subcommand == "select":
            add_select_parser(prefix_parsers, [common])
        elif prefix_subcommand == "query":
            add_query_parser(prefix_parsers, [common])
        elif prefix_subcommand == "use":
            add_use_parser(prefix_parsers, [common])
        elif prefix_subcommand == "info":
            add_info_parser(prefix_parsers, [common])
        elif prefix_subcommand == "validate":
            prefix_parsers.add_parser(
                "validate", help=l10n("validate-help"), parents=[common]
            ).set_defaults(func=validate)
        elif prefix_subcommand == "destroy":
            add_destroy_parser(prefix_parsers, [common])
        elif prefix_subcommand == "run":
            add_run_parser(prefix_parsers, [common])
        elif prefix_subcommand == "cfg-update":
            add_cfg_update_parser(prefix_parsers, [common])
    elif subcommand == "mirror":
        add_mirror_parser(subparsers, [common])
    elif subcommand is not None and subcommand in get_prefixes() and generic:
        prefix = subcommand
        # Add all the parsers. This is relatively slow, but necessary for sphinx-argparse
        prefix_parser = subparsers.add_parser(
            prefix, help=l10n("prefix-help", prefix=prefix), parents=[common]
        )
        prefix_parsers = prefix_parser.add_subparsers(dest="prefix_subcommand_name")
        prefix_parser.set_defaults(func=lambda args: prefix_parser.print_help())
        add_merge_parser(prefix_parsers, [common])
        add_search_parser(prefix_parsers, [common])
        add_select_parser(prefix_parsers, [common])
        add_query_parser(prefix_parsers, [common])
        add_use_parser(prefix_parsers, [common])
        add_info_parser(prefix_parsers, [common])
        prefix_parsers.add_parser(
            "validate", help=l10n("validate-help"), parents=[common]
        ).set_defaults(func=validate)
        add_destroy_parser(prefix_parsers, [common])
        add_run_parser(prefix_parsers, [common])
        add_cfg_update_parser(prefix_parsers, [common])

        add_mirror_parser(subparsers, [common])

    add_init_parser(subparsers, parents=[common])
    add_sync_parser(subparsers, parents=[common])
    return parser


def parse_args(
    subcommand: Optional[str] = None, prefix_subcommand: Optional[str] = None
):
    parser = get_parser(subcommand, prefix_subcommand)

    try:
        import argcomplete  # pylint: disable=import-error

        argcomplete.autocomplete(parser)
    except ModuleNotFoundError:
        pass

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(2)

    if "--ignore-default-opts" in sys.argv:
        args = sys.argv[1:]
    else:
        args = sys.argv[1:] + os.environ.get("OMWMERGE_DEFAULT_OPTS", "").split()

    if subcommand is None:
        return parser, parser.parse_known_args(args)[0]
    else:
        return parser, parser.parse_args(args)


def main():
    os.environ["PYTHONUNBUFFERED"] = "1"

    _, args = parse_args()
    init_logger(args)

    if args.subcommand_name in get_prefixes():
        env.set_prefix(args.subcommand_name)

    # Ensure that we read config entries into os.environ
    get_config()

    refresh_env()

    prefix_subcommand_name = None
    if hasattr(args, "prefix_subcommand_name"):
        prefix_subcommand_name = cast(str, args.prefix_subcommand_name)
    parser, args = parse_args(args.subcommand_name, prefix_subcommand_name)

    if args.no_confirm:
        env.INTERACTIVE = False

    init_logger(args)

    if args.version:
        print(f"Portmod {get_version()}")
        sys.exit(0)

    if not args.subcommand_name and not args.version:
        error(l10n("invalid-cli-help"))
        parser.print_help()
        sys.exit(2)

    try:
        args.func(args)
    except (SandboxedError, InputException, SourceUnfetchable, RemoteHashError) as e:
        # Suppress traces for SandboxedErrors when not in debug mode
        if logging.root.level <= logging.DEBUG:
            traceback.print_exc()
        error("{}".format(e))
        sys.exit(1)
    except Exception as e:
        traceback.print_exc()
        error("{}".format(e))
        sys.exit(1)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from typing import Optional


def vdb_path() -> str:
    return os.path.join(root(), os.environ["VARIABLE_DATA"], "db")


def prefix_name() -> Optional[str]:
    return os.environ.get("PORTMOD_PREFIX_NAME")


def download_dir() -> str:
    return os.environ["PORTMOD_DOWNLOAD_DIR"]


def messages_dir() -> str:
    return os.environ["PORTMOD_MESSAGES_DIR"]


def warnings_dir() -> str:
    return os.environ["PORTMOD_WARNINGS_DIR"]


def root() -> str:
    return os.environ["PORTMOD_ROOT"]


def config_protect_dir() -> str:
    return os.environ["PORTMOD_CONFIG_PROTECT_DIR"]


def tmp_vdb() -> str:
    return os.environ["PORTMOD_TMP_VDB"]


def local_mods() -> str:
    return os.environ["PORTMOD_LOCAL_MODS"]



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI to select various configuration options

Currently just profiles
"""

import os
from logging import info
from typing import List, Optional

from portmod.config import get_config, set_config_value
from portmod.config.profiles import (
    get_profile_name,
    get_profile_path,
    profile_link,
    profile_parents,
    set_profile,
)
from portmod.globals import env
from portmod.modules import add_parsers
from portmod.news import add_news_parsers
from portmod.prompt import display_num_list
from portmod.repo import RemoteRepo, get_repo_name, has_repo
from portmod.repo.metadata import get_profiles
from portmod.repos import add_repo, get_remote_repos, get_repos
from portmod.sync import sync
from portmodlib.colour import bright, green, lblue
from portmodlib.l10n import l10n

from .error import InputException


def list_profiles(arch: str) -> List[str]:
    """
    Prints a list of profiles

    args:
        arch: The architecture to use to filter the profiles

    returns:
        The paths of each profile, in the order they were displayed
    """
    start = 0
    try:
        profile_path: Optional[str] = get_profile_path()
    except FileNotFoundError:
        profile_path = None
    for repo, profiles in get_profiles(arch, env.prefix().REPOS).items():
        print(bright(green(l10n("profile-available", repo=repo))))
        display_num_list(
            [profile for _, profile, _ in profiles],
            [stability for _, _, stability in profiles],
            {
                index + start
                for index, (path, _, _) in enumerate(profiles)
                if path == profile_path
            },
            start=start,
        )
        start += len(profiles)
    return [
        path
        for values in get_profiles(arch, env.prefix().REPOS).values()
        for path, _, _ in values
    ]


def list_repos(arch: str) -> List[RemoteRepo]:
    repos = get_remote_repos(arch).values()
    print(bright(green(l10n("repos-available"))))
    display_num_list(
        [bright(repo.name) + ": " + repo.description for repo in repos],
        [
            repo.quality + (" " + lblue(repo.sync_uri) if repo.sync_uri else "")
            for repo in repos
        ],
        {index for index, repo in enumerate(repos) if has_repo(repo.name)},
    )
    return list(repos)


def add_prefix_repo(repo: str):
    enabled_repos = get_config()["REPOS"]
    repo_name = get_repo_name_from_input(repo)

    if repo_name and repo_name not in enabled_repos:
        enabled_repos.add(repo_name)
        info(l10n("repo-adding", name=repo_name, conf=env.prefix().CONFIG))

    set_config_value("REPOS", " ".join(sorted(enabled_repos)))
    # Re-set prefix so that env.prefix().REPOS is updated
    env.set_prefix(env.PREFIX_NAME)


def get_repo_name_from_input(value: str) -> str:
    repos = list(get_remote_repos(env.prefix().ARCH).values())
    if value.isdigit():
        repo_name = repos[int(value)].name
    else:
        if any(repo.name == value for repo in repos):
            repo_name = value
        else:
            raise InputException(l10n("repo-does-not-exist", name=value))

    return repo_name


def add_profile_parsers(subparsers, parents):
    profile = subparsers.add_parser(
        "profile", help=l10n("profile-help"), parents=parents
    )
    profile_subparsers = profile.add_subparsers()
    profile_list = profile_subparsers.add_parser("list", help=l10n("profile-list-help"))
    profile_set = profile_subparsers.add_parser("set", help=l10n("profile-set-help"))
    profile_set.add_argument(
        "index", metavar=l10n("number-placeholder"), help=l10n("profile-number-help")
    )
    profile_show = profile_subparsers.add_parser("show", help=l10n("profile-show-help"))
    profile_debug = profile_subparsers.add_parser(
        "debug", help=l10n("profile-debug-help")
    )

    def list_func(_args):
        list_profiles(env.prefix().ARCH)

    def set_func(args):
        path, _, _ = [
            tup
            for value in get_profiles(env.prefix().ARCH, env.prefix().REPOS).values()
            for tup in value
        ][int(args.index)]
        set_profile(path)

    def show_func(_args):
        print(bright(green(l10n("profile-current-name", path=profile_link()))))
        try:
            print(
                "  "
                + bright(get_profile_name())
                + f" ({get_repo_name(get_profile_path())})"
            )
        except FileNotFoundError:
            print("  Profile is not set")

    def debug_func(_args):
        print("Full path of profile:")
        print("  " + get_profile_path())

        print("Inherited profile directories:")
        for path in profile_parents():
            print("  ", path)

        print("Files in the profile:")
        for path in profile_parents():
            for file in os.listdir(path):
                if file != "parent":
                    print("  ", os.path.join(path, file))

    def profile_help(args):
        profile.print_help()

    profile.set_defaults(func=profile_help)
    profile_list.set_defaults(func=list_func)
    profile_set.set_defaults(func=set_func)
    profile_show.set_defaults(func=show_func)
    profile_debug.set_defaults(func=debug_func)


def add_repo_parser(subparsers, parents):
    repo = subparsers.add_parser("repo", help=l10n("repo-help"), parents=parents)
    repo_subparsers = repo.add_subparsers()
    repo_list = repo_subparsers.add_parser("list", help=l10n("repo-list-help"))
    repo_add = repo_subparsers.add_parser("add", help=l10n("repo-add-help"))
    repo_add.add_argument(
        "repo", metavar=l10n("repo-placeholder"), help=l10n("repo-identifier-help")
    )
    repo_remove = repo_subparsers.add_parser("remove", help=l10n("repo-remove-help"))
    repo_remove.add_argument(
        "repo", metavar=l10n("repo-placeholder"), help=l10n("repo-identifier-help")
    )

    def list_func(_args):
        list_repos(env.prefix().ARCH)

    def add_func(args):
        add_prefix_repo(args.repo)
        repo = get_remote_repos(env.prefix().ARCH)[args.repo]
        result = add_repo(repo)
        if result:
            sync([result])
        env.REPOS = get_repos()

    def remove_func(args):
        enabled_repos = get_config()["REPOS"]
        repo_name = get_repo_name(args.repo)

        if repo_name and repo_name in enabled_repos:
            enabled_repos.remove(repo_name)
            info(l10n("repo-removing", name=repo_name, conf=env.prefix().CONFIG))

        set_config_value("REPOS", " ".join(sorted(enabled_repos)))

    def repo_help(args):
        repo.print_help()

    repo.set_defaults(func=repo_help)
    repo_list.set_defaults(func=list_func)
    repo_add.set_defaults(func=add_func)
    repo_remove.set_defaults(func=remove_func)


def add_select_parser(subparsers, parents):
    """
    Adds the select subparser to the given subparsers
    """
    parser = subparsers.add_parser("select", help=l10n("select-help"), parents=parents)
    _subparsers = parser.add_subparsers()
    add_profile_parsers(_subparsers, parents)
    add_news_parsers(_subparsers, parents)
    add_repo_parser(_subparsers, parents)
    add_parsers(_subparsers, parents)
    parser.set_defaults(func=lambda args: parser.print_help())



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
The module accessible within pybuilds

Note that this module should not be imported outside of pybuild files
"""

from portmodlib._deprecated import (  # noqa  # pylint: disable=unused-import
    File,
    InstallDir,
)
from portmodlib._deprecated.vfs import (  # noqa  # pylint: disable=unused-import
    find_file,
    list_dir,
)
from portmodlib.atom import version_gt  # noqa  # pylint: disable=unused-import
from portmodlib.fs import patch_dir  # noqa  # pylint: disable=unused-import
from portmodlib.globals import download_dir as _download_dir
from portmodlib.masters import get_masters  # noqa  # pylint: disable=unused-import
from portmodlib.usestr import (  # noqa  # pylint: disable=unused-import
    check_required_use,
    use_reduce,
)

from ._pybuild import (  # noqa  # pylint: disable=unused-import
    Pybuild1,
    Pybuild2,
    apply_patch,
)

DOWNLOAD_DIR = _download_dir()

__all__ = [
    "version_gt",
    "patch_dir",
    "get_masters",
    "use_reduce",
    "find_file",
    "list_dir",
    "File",
    "InstallDir",
    "Pybuild1",
    "Pybuild2",
    "apply_patch",
    "DOWNLOAD_DIR",
]



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the mod selection system
"""

import io
import sys
from contextlib import redirect_stdout
from io import StringIO

import pytest

from portmod._cli.main import main
from portmod.query import display_search_results, query, query_depends
from portmodlib.atom import Atom, atom_sat

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    """
    Sets up test repo for querying
    """
    yield setup_env("test")
    tear_down_env()


def test_query():
    """
    Tests that we can query for exact matches in metadata fields
    """
    results = query("license:eula")
    assert any(pkg.cpn == "test/test-eula" for pkg in results)


def test_insensitive_squelch():
    """
    Tests that we can query for case insensitive matches where there are separators
    in between keywords
    """
    results = query('desc:"desc foo"')
    assert any(pkg.cpn == "test/test" for pkg in results)


def test_depends():
    """
    Tests that we can query for mods that depend on a particular atom
    """
    results = query_depends(Atom("test/test"), all_mods=True)
    assert any(atom_sat(atom, Atom("test/test2-1.0")) for atom, _ in results)


def test_display_results():
    """
    Tests that display_search_results doesn't cause any exceptions
    and that all mods are included in the result
    """
    results = query("test")
    assert results
    strfile = io.StringIO()
    display_search_results(results, file=strfile)
    string = strfile.getvalue()
    for package in results:
        assert package.cpn in string


def test_main_depends():
    """Tests that the cli depends query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "-a", "depends", "test/test5"]
        main()
        assert "test/test7" in output.getvalue()


def test_main_has():
    """Tests that the cli has query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "-a", "has", "DATA_OVERRIDES"]
        main()
        assert "test/test6" in output.getvalue()


def test_main_hasuse():
    """Tests that the cli hasuse query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "-a", "hasuse", "baf"]
        main()
        assert "test/test4" in output.getvalue()


def test_main_uses():
    """Tests that the cli uses query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "uses", "test/test7"]
        main()
        string = output.getvalue()
        assert "foo" in string
        assert "test flag" in string


def test_main_meta():
    """Tests that the cli has query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "meta", "test/test"]
        main()
        assert "someone <someone@example.org>" in output.getvalue()


def test_main_list():
    """Tests that the cli has query interface functions sanely"""
    output = StringIO()
    with redirect_stdout(output):
        sys.argv = ["portmod", "test", "query", "list", "test", "--remote"]
        main()
    assert "test/test-1.0" in output.getvalue()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from colorama import Fore, Style, init

init()


def yellow(text):
    return "{}{}{}".format(Fore.YELLOW, text, Style.RESET_ALL)


def blue(text):
    return "{}{}{}".format(Fore.BLUE, text, Style.RESET_ALL)


def lcyan(text):
    return "{}{}{}".format(Fore.LIGHTCYAN_EX, text, Style.RESET_ALL)


def cyan(text):
    return "{}{}{}".format(Fore.CYAN, text, Style.RESET_ALL)


def lblue(text):
    return "{}{}{}".format(Fore.LIGHTBLUE_EX, text, Style.RESET_ALL)


def green(text):
    return "{}{}{}".format(Fore.GREEN, text, Style.RESET_ALL)


def lred(text):
    return "{}{}{}".format(Fore.LIGHTRED_EX, text, Style.RESET_ALL)


def red(text):
    return "{}{}{}".format(Fore.RED, text, Style.RESET_ALL)


def magenta(text):
    return "{}{}{}".format(Fore.MAGENTA, text, Style.RESET_ALL)


def lgreen(text):
    return "{}{}{}".format(Fore.LIGHTGREEN_EX, text, Style.RESET_ALL)


def bright(text):
    return "{}{}{}".format(Style.BRIGHT, text, Style.RESET_ALL)


def colour(colour, text):
    return "{}{}{}".format(colour, text, Style.RESET_ALL)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests common/* packages
"""

import pytest

from portmod.merge import configure

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    yield setup_env("test")
    tear_down_env()


def test_common_install():
    """Tests that packages with common dependencies install correctly"""
    configure(["test/test-new"])
    configure(["test/test-new"], depclean=True)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Dependency resolution module

Converts mod dependency links and REQUIRED_USE conditions into a MAX-SAT formula in
conjunctive normal form.
This formula is then solved using pysat (python-sat on pypi) and the resulting model
converted back into a list of installed mods and their use flag configuration.

Note that the hard requirements defined in DEPEND, RDEPEND and REQUIRED_USE are
converted into a SAT formula that must be solved in its entirety.
We use a MAX-SAT solver because there are also other soft requirements which are used
to avoid installing mods unnecessarily and to avoid changing the user's use flag
configuration, if possible.

See https://en.wikipedia.org/wiki/Boolean_satisfiability_problem for details on
the SAT problem
"""

from collections import defaultdict
from logging import info, warning
from typing import AbstractSet, Dict, Iterable, List, Optional, Set, Tuple

from portmod.config.sets import get_set
from portmod.config.use import get_use
from portmod.globals import env
from portmod.loader import (
    AmbiguousAtom,
    load_all_installed,
    load_installed_pkg,
    load_pkg,
    load_pkg_fq,
)
from portmod.transactions import (
    PackageDoesNotExist,
    Transactions,
    UseDep,
    generate_transactions,
)
from portmodlib.atom import Atom, FQAtom, atom_sat
from portmodlib.l10n import l10n

from .formula import Formula, Target, add_masked, cond_flagstr, fstr, generate_formula
from .tokens import AtomToken, FlagToken, Token, VariableToken
from .weights import weigh_clauses


class DepError(Exception):
    """Indicates an unsatisfiable transaction"""


class _HiddenToken(str):
    def __str__(self):
        return ""


def find_dependency(
    clauses: Iterable[Formula.Clause], model: Set[Token], modstr: str
) -> Optional[Formula.Clause]:
    """Finds clause that depends on the given atom"""
    # Ignore top-level dependencies
    if not isinstance(modstr, Atom):
        return None

    for clause in clauses:
        if isinstance(clause, Formula.DepClause):
            if atom_sat(Atom(modstr), clause.dependency) and all(
                var in model for var in clause.requirements
            ):
                return clause
    return None


def find_conflict(
    clauses: Iterable[Formula.Clause], display_conflict: bool = False
) -> Tuple[List[Formula.Clause], Optional[List[Formula.Clause]]]:
    """
    Produces traces for the provided conflicting clauses

    Find clause that caused the solver to fail

    Then backtrack the package that added that clause until we reach @selected by
    looking for clauses containing the (enabled) mod as a token.
    Returns this trace.

    If display_conflict is True, also return trace for a package that requires/blocks
    the clause that caused the failure

    args:
        clauses: The clauses to be solved. These are assumed to contain a conflict
        display_conflict: If true, return a second list describing the package which
                          requires/blocks the package which caused the failure
                          Otherwise, the second list will always be None
    raises:
        DepError: if no conflict was found in the clauses

    returns:
        The conflicting clause and its parent clauses.
    """
    from pysat.solvers import Solver

    # Note: we use mc by itself rather than with RC2 (as used in the main resolve function)
    # since we don't need to take into account weighted "soft" clauses.
    with Solver("mc") as solver:
        solveableformula = []
        # Add atmost clauses first, as they won't by themselves cause conflicts,
        # and are not very useful for explaining a failed transaction
        for clause in clauses:
            if isinstance(clause, Formula.MetaClause) and clause.atmost is not None:
                solver.add_atmost(clause.intclause, clause.atmost)

        lastmodel = set()
        # FIXME: It's more efficient to do a binary rather than linear search
        for clause in clauses:
            # Ignore atmost clauses we already added, and weighted clauses
            # which will never cause a conflict
            if isinstance(clause, Formula.MetaClause) and (
                clause.atmost is not None or clause.weight is not None
            ):
                continue

            solver.add_clause(clause.intclause)

            if solver.solve():
                lastmodel = solver.get_model()
                solveableformula.append(clause)
            else:
                conflict = None
                model = set(map(Formula.getstring, lastmodel))

                for solveableclause in solveableformula:
                    # Find clause that contradicts failed clause
                    # Note that metaclauses don't have a blocks function,
                    if solveableclause.blocks(model, clause):
                        conflict = solveableclause
                        break

                source_trace = [clause]
                conflict_trace = None
                parent = find_dependency(solveableformula, model, clause.source)
                i = 1
                # Note: Any descriptor which is empty (when stringified) will be ignored
                #       This is for clauses added for constructed traces such as the use
                #       flag dependency traces
                while parent is not None and str(parent.source):
                    source_trace.append(parent)
                    parent = find_dependency(solveableformula, model, parent.source)
                    i += 1

                if conflict and display_conflict:
                    conflict_trace = [conflict]
                    parent = find_dependency(solveableformula, model, conflict.source)
                    i = 1
                    while parent is not None and str(parent.source):
                        conflict_trace.append(parent)
                        parent = find_dependency(solveableformula, model, parent.source)
                        i += 1

                return source_trace, conflict_trace
    raise DepError("Internal error: Unable to find conflict!")


def resolve(
    enabled: Iterable[Atom],
    disabled: Iterable[Atom],
    explicit: AbstractSet[Atom],
    selected: AbstractSet[Atom],
    selected_sets: AbstractSet[str],
    *,
    deep: bool = False,
    update: bool = False,
    depclean: bool = False,
    emptytree: bool = False,
) -> Transactions:
    """
    Calculates new mod configuration to match system after the given mods are installed

    Note: We have two modes of operation:

    Shallow
        We assume that all installed mods are fixed and will not
        change version. Any version of a newly selected mods may be installed.
        Note that use flags may change on installed mods.
    Deep
        We treat every mod as newly selected, and choose from among its versions

    args:
        enabled: Packages which are to be enabled/installed
        disabled: Packages which are to be disabled/removed
        selected: Enabled packages which were explicitly selected
        selected_sets: Sets which were explicitly selected (the contents of which
                would be in enabled)
        deep: Whether or not we are running in deep mode
        update: If true, packages will be updated, if possible.
        depclean: If true, packages which were neither explicitly selected, nor
                  required as dependencies, will be removed.
        emptytree: If true, all packages in the dependency tree will be rebuilt, as
                   if nothing was installed.
    returns:
        Transactions object representing the package changes required
    """
    # Slow imports
    from pysat.examples.rc2 import RC2

    info(l10n("calculating-dependencies"))
    formula = Formula()

    # List of sets of mod objects, with each being a specific version of that mod
    oldselected: List[Target] = []
    newenabled: Dict[str, Target] = dict()

    CMD_ATOM = "packages passed on command line"
    WORLD_ATOM = "world favourites file"

    for atom in list(enabled) + list(disabled):
        if not load_pkg(atom):
            raise PackageDoesNotExist(atom)

    newenabledset = {
        atom: CMD_ATOM
        for atom in set(enabled)
        | {atom for set_name in selected_sets for atom in get_set(set_name)}
    }
    for atom in disabled:
        name = load_pkg(atom)[0].CPN
        if name in newenabledset:
            del newenabledset[name]

    for atom in disabled:
        for mod in load_pkg(atom):
            formula.append_dep([AtomToken(mod.ATOM, polarity=False)], CMD_ATOM, atom)

    def create_modlist(atom):
        modlist = load_pkg(atom)

        # Raise exception if mod name is ambiguous (exists in multiple categories)
        if not all(mod.ATOM.C == modlist[0].ATOM.C for mod in modlist):
            raise AmbiguousAtom(atom, {mod.CPN for mod in modlist})

        if not modlist:
            if atom in set(selected):
                raise PackageDoesNotExist(atom)

            raise PackageDoesNotExist(
                msg=l10n("package-does-not-exist-in-world", atom=atom)
            )
        return modlist

    for atom, source in newenabledset.items():
        modlist = create_modlist(atom)
        name = modlist[0].CPN
        if name in newenabled:
            newenabled[modlist[0].CPN].pkgs.extend(modlist)
            if newenabled[modlist[0].CPN].source is CMD_ATOM or source is CMD_ATOM:
                # Use generic atom if included multiple times on command line.
                # Not all versions in modlist will correspond to a specific version
                # passed on the command line.
                newenabled[modlist[0].CPN].atom = name
            # Prefer command line as source rather than world file
            if newenabled[modlist[0].CPN].source is WORLD_ATOM and source is CMD_ATOM:
                newenabled[modlist[0].CPN].source = CMD_ATOM
        else:
            newenabled[modlist[0].CPN] = Target(modlist, atom, source)

    for atom in get_set("world") - {load_pkg(atom)[0].CPN for atom in disabled}:
        if atom not in selected:
            modlist = create_modlist(atom)
            oldselected.append(Target(modlist, atom, "world favourites file"))

    # Any remaining installed mods don't need to remain installed if there aren't
    # any dependencies, so source is None
    installed = [Target([mod], mod.ATOM, None) for mod in load_all_installed()]

    selected_cpn = set()
    explicit_cpn = set()

    for atom in explicit:
        pkg = load_pkg(atom)[0]
        explicit_cpn.add(pkg.CPN)

    for atom in selected:
        pkg = load_pkg(atom)[0]
        selected_cpn.add(pkg.CPN)

    # Hard clauses

    # The explicitly passed packages are aleays handled in deep mode
    formula.merge(generate_formula(list(newenabled.values()), set(), deep=True))

    # Existing packages are assumed not to need to change if not in deep mode,
    # so we ignore their dependency tree (unless run in deep mode).
    formula.merge(generate_formula(installed, set(), deep=deep))
    for target in oldselected:
        if load_installed_pkg(target.atom):
            formula.merge(generate_formula([target], set(), deep=deep))
        else:
            # world packages which are not installed should be merged in deep mode,
            # as they may need extra dependencies
            formula.merge(generate_formula([target], set(), deep=True))

    # Soft clauses
    formula.merge(
        weigh_clauses(
            formula.atoms,
            formula.flags,
            explicit=explicit_cpn,
            deep=deep,
            depclean=depclean,
            update=update,
            emptytree=emptytree,
        )
    )

    formula.merge(add_masked(formula.atoms, formula.flags))

    if depclean:
        for pkg in load_all_installed():
            # When depcleaning, installed packages should be frozen at their current
            # version, or else removed
            versions = load_pkg(Atom(pkg.ATOM.CPN))
            var = formula.genvariable(
                [f"No non-installed versions of {pkg.CPN} are allowed"]
            )
            # Clause requires that either the installed version should be removed,
            # or all other versions should be removed.
            # This means it is not possible for a version other than the installed version
            # to be kept
            formula.append_dep(
                [AtomToken(pkg.ATOM, polarity=False), var],
                "Selected packages must not be changed when depcleaning",
                pkg.ATOM,
            )
            # Var can only be true if all versions are not installed
            # as each clause requires that either a particular version is not installed,
            # or var is false.
            # Hence if var is true, every one of these clauses is satisfied by the left term
            for other in versions:
                if not other.INSTALLED:
                    formula.append(
                        [AtomToken(other.ATOM, polarity=False), var.neg()],
                        other.ATOM,
                        "Selected packages must not be changed when depcleaning",
                    )

            for flag in pkg.IUSE_EFFECTIVE:
                if flag in get_use(pkg)[0]:
                    flagdep = fstr(pkg.ATOM, flag)
                else:
                    flagdep = fstr(pkg.ATOM, flag).neg()
                formula.append_usedep(
                    [flagdep, AtomToken(pkg.ATOM, polarity=False)],
                    "Selected packages must not be changed when depcleaning",
                    pkg.ATOM,
                    flagdep,
                )

    if not formula.clauses:
        return Transactions()

    formula.make_numeric()
    wcnf = formula.get_wcnfplus()
    solver = RC2(wcnf, solver="mc")
    solver.compute()
    if solver.compute():
        info(l10n("done"))
        # Turn numbers in result back into strings
        result = set(
            filter(
                # Filter out custom variables that are only meaningful
                # for the computation
                lambda x: not isinstance(x, VariableToken),
                [Formula.getstring(num) for num in solver.model],
            )
        )
        flags = [token for token in result if isinstance(token, FlagToken)]
        enabled_final = [
            token.value
            for token in result
            if token.polarity and isinstance(token, AtomToken)
        ]
        enablednames = [atom.CPN for atom in enabled_final]
        disabled_final = [
            FQAtom(token.value)
            for token in result
            if not token.polarity and isinstance(token, AtomToken)
            # If mod is enabled and installed version is disabled,
            # ignore disabled version, and vice versa
            and token.value.CPN not in enablednames
        ]

        flag_dict: Dict[FQAtom, Set[str]] = defaultdict(set)
        for token in flags:
            if token.value.strip_use() in enabled_final:
                flag_dict[token.value.strip_use()].add(str(token))

        def is_installed(atom: Atom) -> bool:
            for fqatom in enabled_final:
                if atom_sat(fqatom, atom) and atom.USE <= flag_dict[fqatom]:
                    return True
            return False

        usedeps = []

        for token in flags:
            atom = token.value.strip_use()
            flag = list(token.value.USE)[0]
            if token.polarity:
                prefix = ""
            else:
                prefix = "-"

            # For all changes to the flag configuration, determine what caused the
            # change by re-running the solver with the requirement that the inverse of
            # the flag is set, and finding the conflicting clause
            if atom in enabled_final:
                pkg = load_pkg_fq(atom)
                # Note: this matches the condition in _deps.weights and is necessary since we
                # deliberately freeze flags on packages not explicitly specified
                if (
                    pkg.INSTALLED
                    and (deep and update or pkg.CPN in explicit_cpn)
                    or not pkg.INSTALLED
                ):
                    enabled_use, _ = get_use(pkg, is_installed=is_installed)
                else:
                    enabled_use = pkg.get_use()
                if (
                    not token.polarity
                    and flag in enabled_use
                    or token.polarity
                    and flag not in enabled_use
                ):
                    # To prevent solver from enabling a different version of the package,
                    # also require that the package is enabled
                    flag_config = [
                        Formula.UseDepClause(
                            _HiddenToken(), [token.neg()], atom, token.neg()
                        ).str2num(),
                        Formula.DepClause(
                            _HiddenToken(), [AtomToken(atom, True)], atom
                        ).str2num(),
                    ]

                    # If this is a result of a required use when a different flag is set,
                    # e.g. foo? ( bar ), then we need to force the rest of the configuration
                    for otherflag in pkg.IUSE_EFFECTIVE:
                        if flag != otherflag:
                            newtoken = cond_flagstr(atom, otherflag)
                            if otherflag not in enabled_use:
                                newtoken = newtoken.neg()
                            flag_config.append(
                                Formula.UseDepClause(
                                    _HiddenToken(), [newtoken], atom, newtoken
                                ).str2num(),
                            )

                    comment = []
                    try:
                        source_trace, _ = find_conflict(flag_config + formula.clauses)
                        for index, clause in enumerate(source_trace):
                            comment.append("# " + index * "  " + clause.colourless())
                    except DepError as error:
                        message = f"Internal Error: Unable to trace flag {flag} for package {pkg}"
                        if env.TESTING:
                            raise Exception(message) from error
                        warning(message)

                    usedeps.append(UseDep(atom, prefix + flag, None, comment))

        transactions = generate_transactions(
            enabled_final,
            disabled_final,
            selected_cpn,
            usedeps,
            flag_dict,
            emptytree=emptytree,
            update=update,
        )
        return transactions

    source_trace, conflict_trace = find_conflict(formula.clauses, display_conflict=True)
    exceptionstring = ""
    for index, clause in enumerate(source_trace):
        exceptionstring += index * "  " + f"{clause}\n"

    # conflict_trace may be empty if the clause contradicts itself
    if conflict_trace:
        exceptionstring += l10n("contradicts") + "\n"

        for index, clause in enumerate(conflict_trace):
            exceptionstring += index * "  " + f"{clause}\n"

    raise DepError(l10n("unable-to-satisfy-dependencies") + f"\n{exceptionstring}")



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for reading from list files"""

import os
from typing import List, Optional


class CommentedLine(str):
    """A line in a list file which is preceeded by a comment"""

    comment: List[str]


def read_list(listpath: str, encoding: Optional[str] = None) -> List[str]:
    """
    Reads the given list file and returns its contents

    Comments are supported, and comment blocks get attached to all lines following them
    until a blank line is encountered.
    """
    results: List[str] = []
    comment: Optional[List[str]] = None
    with open(listpath, mode="r", encoding=encoding) as list_file:
        for line in list_file.read().splitlines():
            line = line.strip()
            if line.startswith("#"):
                if comment:
                    comment.append(line)
                else:
                    comment = [line]
            elif line:
                if comment:
                    commented = CommentedLine(line)
                    commented.comment = comment
                    results.append(commented)
                else:
                    results.append(line)
            else:
                comment = None
    return results


def add_list(listpath: str, entry: str):
    """Appends the given value to the list file"""
    os.makedirs(os.path.dirname(listpath), exist_ok=True)
    with open(listpath, mode="a") as list_file:
        print(entry, file=list_file)


def write_list(path: str, contents: List[str]):
    """Writes a list file containing the given list"""
    with open(path, "w") as file:
        for value in contents:
            print(value, file=file)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Atom parsing tests
"""

import pytest

from portmodlib.atom import Atom, InvalidAtom


def test_invalid():
    with pytest.raises(InvalidAtom):
        Atom("foo/bar baz")


def test_valid():
    Atom("foo/bar-baz-1.0.20.3-r123::foo[-a,b?,!c?,c]")


def test_complex_version():
    Atom("foo/bar-1.0a_alpha")
    Atom("foo/bar-1.0a_alpha12")
    Atom("foo/bar-1.0a_pre1")
    Atom("foo/bar-1.0a_beta2")
    Atom("foo/bar-1.0a_rc1")
    Atom("foo/bar-1.0a_p1")
    Atom("foo/bar-2-1.0a_p1")


def test_nodots():
    assert Atom("foo/bar-1").PV == "1"


def test_epoch():
    atom = Atom("foo-e1-1.0.0-r1")
    assert atom.version
    assert atom.version.numeric == ["1", "0", "0"]
    assert atom.version.epoch == 1
    assert atom.version.revision == 1
    assert atom.version.display(revision=True, epoch=False) == "1.0.0-r1"
    assert atom.version.display(revision=False, epoch=True) == "e1-1.0.0"
    assert atom.version.display(revision=False, epoch=False) == "1.0.0"



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests support for installing on top of existing data
"""

import os
import sys

import pytest

from portmod._cli.main import main
from portmod._cli.select import add_prefix_repo
from portmod.config import get_config
from portmod.globals import env
from portmod.merge import configure
from portmodlib.atom import Atom

from .env import select_profile, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    yield setup_env("base")
    tear_down_env()


def run(args):
    sys.argv = args
    main()


def test_init_destroy():
    """
    Tests that prefixes created on top of existing directories
    will not remove existing files on destruction
    """
    with open("existing_file", "w") as file:
        file.write("")

    run(["portmod", "init", "test2", "test", "."])
    run(["portmod", "test2", "destroy"])
    assert os.path.exists("existing_file")


def test_init_overlap():
    """Tests that you cannot create two prefixess with overlapping directories"""
    run(["portmod", "init", "test2", "test", "."])
    os.makedirs("foo", exist_ok=True)
    with pytest.raises(SystemExit):
        run(["portmod", "init", "test3", "test", "foo"])

    run(["portmod", "test2", "destroy"])

    run(["portmod", "init", "test2", "test", "foo"])
    with pytest.raises(SystemExit):
        run(["portmod", "init", "test3", "test", "."])

    run(["portmod", "test2", "destroy"])


@pytest.mark.skipif(
    sys.platform == "win32" and "APPVEYOR" in os.environ,
    reason="Windows CI is flaky with deleting git repositories",
)
def test_backup():
    """
    Tests that when packages overwrite existing data,
    it's restored after the package is removed
    """
    filepath = os.path.join("pkg", "test", "test", "Foo")
    os.makedirs(os.path.dirname(filepath), exist_ok=True)
    with open(filepath, "w") as file:
        file.write("bar")

    run(["portmod", "init", "test2", "test", "."])
    env.set_prefix("test2")
    add_prefix_repo("test")
    env.set_prefix("test2")
    select_profile("test")
    get_config.cache_clear()

    configure([Atom("=test/test-1.0-r2")])
    with open(filepath) as file:
        assert file.read() != "bar"
    configure([Atom("test/test")], depclean=True)
    with open(filepath) as file:
        assert file.read() == "bar"

    run(["portmod", "test2", "destroy"])



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
import stat
from fnmatch import fnmatch, fnmatchcase
from functools import lru_cache
from pathlib import Path
from shutil import copystat
from typing import Callable, Generator, List, Optional, Set, Tuple, Union

from portmodlib.portmod import _get_hash
from portmodlib.source import HashAlg

# 32MB buffer seems to give the best balance between performance on large files
# and on small files
HASH_BUF_SIZE = 32 * 1024 * 1024

try:
    DirEntryStr = os.DirEntry[str]
except TypeError:
    # Python 3.7 and 3.8 don't like subscripting os.DirEntry
    DirEntryStr = os.DirEntry  # type: ignore


def onerror(func, path, _exc_info):
    """
    Error handler for ``shutil.rmtree``.

    If the error is due to an access error (read only file)
    it attempts to add write permission and then retries.

    If the error is for another reason it re-raises the error.

    Usage : ``shutil.rmtree(path, onerror=onerror)``
    """
    if not os.access(path, os.W_OK):
        # Is the error an access error ?
        os.chmod(path, stat.S_IWUSR)
        func(path)
    else:
        raise  # pylint: disable=misplaced-bare-raise


def _move2(src: DirEntryStr, dest: str):
    if os.path.islink(src):
        return os.symlink(os.readlink(src.path), dest)
    return shutil.move(src.path, dest)


def _patch_file(
    src: os.DirEntry,
    dst: str,
    overwrite: bool = True,
    move_function: Callable[[os.DirEntry, str], None] = _move2,
):
    if os.path.exists(dst) and src.is_file():
        if overwrite:
            os.remove(dst)
        else:
            raise FileExistsError(f"File {dst} already exists")

    move_function(src, dst)


def _iter_files_to_patch(
    src: Union[str, DirEntryStr],
    dst: str,
    *,
    ignore: Optional[Callable[[str, List[str]], Set[str]]] = None,
    case_sensitive: bool = True,
) -> Generator[Tuple[os.DirEntry, str], None, None]:
    with os.scandir(src) as itr:
        entries = list(itr)
    if ignore is not None:
        ignored_names = ignore(os.fspath(src), [x.name for x in entries])
    else:
        ignored_names = set()

    for entry in entries:
        if entry.name in ignored_names:
            continue
        if case_sensitive:
            dstname = os.path.join(dst, entry.name)
        else:
            dstname = ci_exists(os.path.join(dst, entry.name)) or os.path.join(
                dst, entry.name
            )

        if entry.is_symlink():
            yield (entry, dstname)
        elif entry.is_dir():
            yield from _iter_files_to_patch(
                entry,
                dstname,
                ignore=ignore,
                case_sensitive=case_sensitive,
            )
        else:
            yield (entry, dstname)


# Modified version of shutil.copytree from
# https://github.com/python/cpython
# Python software and documentation are licensed under the
# Python Software Foundation License Version 2
def patch_dir(
    src: Union[str, os.DirEntry],
    dst: str,
    *,
    overwrite: bool = True,
    ignore: Optional[Callable[[str, List[str]], Set[str]]] = None,
    case_sensitive: bool = True,
    move_function: Callable[[os.DirEntry, str], None] = _move2,
) -> str:
    """
    Copies src ontop of dst

    args:
        src: Source directory to copy from
        dst: Destination directory to copy to
        overwrite: If true, overwrite existing files.
        ignore: A callable which, given a directory and its contents, should return
            a set of files to ignore
        case_sensitive: If False, treat file and directory names as case insensitive
        move_function: The function to use to transfer individual files.
            Default is shutil.move (modified to accept a DirEntry).
            The signature should match shutil.copy2.

    raises:
        FileExistsError

    returns:
        Returns dst
    """

    for src_file, dst_file in _iter_files_to_patch(
        src,
        dst,
        ignore=ignore,
        case_sensitive=case_sensitive,
    ):
        parent_dir = os.path.dirname(dst_file)
        if not os.path.isdir(parent_dir):
            os.makedirs(parent_dir)
            try:
                copystat(os.path.dirname(src_file.path), parent_dir)
            except OSError as why:
                if getattr(why, "winerror", None) is None:
                    raise why

        _patch_file(
            src_file,
            dst_file,
            overwrite=overwrite,
            move_function=move_function,
        )

    return dst


def ci_exists(path: str, *, prefix: Optional[str] = None) -> Optional[str]:
    """
    Checks if a path exists, ignoring case.

    If the path exists but is ambiguous the result is not guaranteed

    args:
        path: The path to check. This path must either be absolute, or be relative to the prefix
        prefix: A leading path to ignore. If path is relative, it is treated as relative to this directory
                If prefix is None, path is treated as relative to the current working directory.

                Case-insensitive checks will not be performed on the components of the prefix
    """
    if os.path.isabs(path) and os.path.exists(path):
        return path

    if os.path.isabs(path):
        partial_path = prefix or "/"
    else:
        partial_path = prefix or os.getcwd()
        if os.path.exists(os.path.join(partial_path, path)):
            return os.path.join(partial_path, path)

    if os.path.isabs(path):
        # Ignore leading empty component when splitting absolute paths
        components = os.path.normpath(path).split(os.sep)[1:]
    else:
        components = os.path.normpath(path).split(os.sep)

    for component in components:
        found = False
        # The entry that exists is not a directory, so it cannot have any contents
        if not os.path.isdir(partial_path):
            return None
        for entryname in os.listdir(partial_path):
            if entryname.lower() == component.lower():
                partial_path = os.path.join(partial_path, entryname)
                found = True
                break
        if not found:
            return None

    if os.path.exists(partial_path):
        return partial_path

    return None


def get_tree_size(path):
    """Return total size of files in given path and subdirs."""
    total = 0
    for entry in os.scandir(path):
        if entry.is_dir(follow_symlinks=False):
            total += get_tree_size(entry.path)
        else:
            total += entry.stat(follow_symlinks=False).st_size
    return total


@lru_cache(maxsize=None)
def get_hash(filename: str, funcs=(HashAlg.BLAKE3,)) -> List[str]:
    """Hashes the given file"""
    return _get_hash(filename, [func.value for func in funcs], HASH_BUF_SIZE)


def is_parent(path: str, prefix: str) -> bool:
    """
    Returns true if and only if prefix is a parent directory of path

    args:
        path: An absolute path
        prefix: An absolute path
    returns:
        True if and only if prefix is a parent directory of path
    """
    path = os.path.normpath(os.path.abspath(path))
    prefix = os.path.normpath(os.path.abspath(prefix))
    return (
        os.path.splitdrive(path)[0] == os.path.splitdrive(prefix)[0]
        and os.path.commonpath([path, prefix]) == prefix
    )


def make_unique_filename(path: str, *, case_insensitive: bool = False) -> str:
    numeric_component = 1
    new_name = os.path.basename(path)
    directory = os.path.dirname(path)

    while (case_insensitive and ci_exists(new_name, prefix=directory)) or (
        not case_insensitive and os.path.exists(os.path.join(directory, new_name))
    ):
        name, ext = os.path.splitext(os.path.basename(path))
        new_name = f"{name}.{numeric_component}{ext}"

    return os.path.join(directory, new_name)


def match(path: Path, pattern: str) -> bool:
    """
    Returns true if the glob-style pattern matches the given path, relative to root

    The pattern can contain :py:mod:`fnmatch`-style patterns
    however it breaks them up per path component, and ``**`` can be used to
    match recursively.

    args:
        path: a relative path
        pattern: A relative glob-style pattern to match path to.
    returns:
        True if and only if path matches pattern.
    """
    assert not path.is_absolute()

    pattern_path = Path(pattern)

    path_parts = list(reversed(path.parts))
    pattern_parts = list(reversed(pattern_path.parts))

    def match_inner(path_parts: List[str], pattern_parts: List[str]) -> bool:
        while path_parts and pattern_parts:
            path_part = path_parts.pop()
            pattern_part = pattern_parts.pop()

            if pattern_part == "**":
                # Should match all remaining path parts
                # Base case: Match ** against nothing.
                # Otherwise, re-insert it into the pattern and try on the remaining path
                return match_inner(
                    path_parts + [path_part], list(pattern_parts)
                ) or match_inner(list(path_parts), pattern_parts + [pattern_part])

            if os.environ.get("CASE_INSENSITIVE_FILES"):
                if not fnmatch(path_part, pattern_part):
                    return False
            else:
                if not fnmatchcase(path_part, pattern_part):
                    return False
        # Only match if there are no components left
        return not path_parts and not pattern_parts

    return match_inner(path_parts, pattern_parts)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
from logging import error, info, warning
from typing import Iterable, cast

from packaging import version

from portmod.globals import env, get_version
from portmod.news import update_news
from portmod.query import update_index
from portmod.repo import Repo, get_repo, has_repo
from portmod.repo.metadata import get_master_names
from portmod.repos import add_repo, get_local_repos, get_remote_repos
from portmodlib.l10n import l10n


def sync(repos: Iterable[Repo]):
    old_prefix = env.set_prefix(None)
    # Slow imports
    import git

    to_sync = list(repos)

    for repo in to_sync:
        if repo.auto_sync and repo.sync_type == "git" and repo.sync_uri:

            if os.path.exists(repo.location):
                info(l10n("syncing-repo", repo=repo.name))
                gitrepo = git.Repo.init(repo.location)
                current = gitrepo.head.commit

                # Remote location has changed. Update gitrepo to match
                if gitrepo.remotes.origin.url != repo.sync_uri:
                    gitrepo.remotes.origin.set_url(repo.sync_uri)

                gitrepo.remotes.origin.pull(rebase=True)

                for diff in current.diff("HEAD"):
                    if diff.renamed_file:
                        print(
                            "{} {} -> {}".format(
                                diff.change_type, diff.a_path, diff.b_path
                            )
                        )
                    if diff.deleted_file:
                        print("{} {}".format(diff.change_type, diff.a_path))
                        if diff.a_path.endswith(".pybuild"):
                            # Remove from pybuild cache
                            parts = diff.a_path.split("/")
                            category = parts[0]
                            file = parts[-1].lstrip(".pybuild")
                            path = os.path.join(
                                env.PYBUILD_CACHE_DIR, repo.name, category, file
                            )
                            if os.path.exists(path):
                                os.remove(path)
                    else:
                        print("{} {}".format(diff.change_type, diff.b_path))

                tags = []
                for tag in gitrepo.tags:
                    # Valid tags must have the tag commit be the merge base
                    # A merge base further back indicates a branch point
                    if tag.name.startswith("portmod_v"):
                        base = gitrepo.merge_base(gitrepo.head.commit, tag.commit)
                        if base and base[0] == tag.commit:
                            tags.append(tag)

                this_version = version.parse(get_version())
                newest = max(
                    [version.parse(tag.name.lstrip("portmod_v")) for tag in tags]
                    + [this_version]
                )
                if newest != this_version and not env.TESTING:
                    warning(l10n("update-message"))
                    warning(l10n("current-version", version=this_version))
                    warning(l10n("new-version", version=newest))
                info(l10n("done-syncing-repo", repo=repo.name))
            else:
                git.Repo.clone_from(repo.sync_uri, repo.location)
                info(l10n("initialized-repository", repo=repo.name))
        elif repo.auto_sync:
            error(
                l10n(
                    "invalid-sync-type",
                    type=repo.sync_type,
                    repo=repo.name,
                    supported="git",
                )
            )

        remote_repos = get_remote_repos()
        for master in get_master_names(repo.location):
            if has_repo(master):
                master_repo = get_repo(master)
            else:
                # As the repository did not previously exist, add_repo will not
                # return None
                master_repo = cast(Repo, add_repo(remote_repos[master]))
            if master_repo not in to_sync:
                to_sync.append(master_repo)

    if os.path.exists(env.PYBUILD_CACHE_DIR):
        local = get_local_repos()
        for repo_name in os.listdir(env.PYBUILD_CACHE_DIR):
            path = os.path.join(env.PYBUILD_CACHE_DIR, repo_name)
            if repo_name not in local:
                print(l10n("cache-cleanup", repo=repo_name))
                shutil.rmtree(path)

    update_news()
    env.set_prefix(old_prefix)
    update_index()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import tempfile
from functools import lru_cache
from typing import Any, List, Optional

import portmod
from portmodlib.l10n import l10n
from portmodlib.portmod import directories


@lru_cache()
def get_version() -> str:
    """Returns portmod version"""
    try:
        from ._version import version

        return str(version)
    except ImportError:
        osp = os.path
        possible_root = osp.dirname(osp.dirname(osp.realpath(__file__)))
        if osp.isfile(osp.join(possible_root, ".portmod_not_installed")):
            # Only a dev dependency and is otherwise unneeded at runtime
            import setuptools_scm

            return str(setuptools_scm.get_version(possible_root))
        else:
            try:
                from importlib.metadata import version as version_func
            except ModuleNotFoundError:
                from importlib_metadata import version as version_func  # type: ignore

            return str(version_func("portmod"))


def get_authors() -> List[str]:
    import git

    repo = os.path.dirname(os.path.dirname(__file__))
    if os.path.exists(os.path.join(repo, ".git")):
        gitrepo = git.Repo.init(repo)
        results = []
        for line in gitrepo.git.shortlog(
            "v1.0.1..", numbered=True, summary=True
        ).splitlines():
            _commits, author = line.strip().split("\t")
            results.append(author)
        return results

    return ["Portmod Authors"]


class Env:
    ALLOW_LOAD_ERROR = True
    DEBUG = False
    DATA_DIR: str = directories.data_dir
    CONFIG_DIR: str = directories.config_dir
    CACHE_DIR: str = directories.cache_dir
    INTERACTIVE = True
    TESTING = False
    PREFIX_NAME: Optional[str] = None
    REPOS: List["portmod.repo.Repo"]

    # The following variables are none if a prefix has not been selected
    class Prefix:
        REPOS: List["portmod.repo.Repo"]

        def __init__(self, prefix_name: str):
            from portmod.prefix import get_prefixes

            if prefix_name not in get_prefixes():
                raise Exception(l10n("invalid-prefix", prefix=prefix_name))

            prefix = get_prefixes()[prefix_name]

            self.ARCH: str = prefix.arch
            self.CONFIG_DIR: str = os.path.join(env.CONFIG_DIR, prefix_name)
            self.SET_DIR: str = os.path.join(self.CONFIG_DIR, "sets")
            self.CONFIG: str = os.path.join(self.CONFIG_DIR, "portmod.conf")

            self.ROOT: str = os.path.realpath(
                prefix.path or os.path.join(env.DATA_DIR, prefix_name)
            )
            self.CACHE_DIR = os.path.join(env.CACHE_DIR, "prefix", prefix_name)
            self.PYBUILD_INSTALLED_CACHE = os.path.join(self.CACHE_DIR, "pybuild")
            self.CONFIG_PROTECT_DIR = os.path.join(self.CACHE_DIR, "cfg_protect")
            self.LOCAL_MODS = os.path.join(self.ROOT, "local")
            self.REPOS = []

        def __setattr__(self, name: str, value: Any):
            if isinstance(value, str):
                os.environ["PORTMOD_" + name] = value
            super().__setattr__(name, value)

    def prefix(self) -> Prefix:
        if self.PREFIX_NAME is not None:
            return self._prefix
        raise Exception("Internal error: Prefix has not been initialized!")

    def set_prefix(self, prefix: Optional[str]) -> Optional[str]:
        from portmod.lock import has_prefix_exclusive_lock, has_prefix_lock
        from portmod.repos import get_repos

        OLD_PREFIX_NAME = self.PREFIX_NAME
        if self.PREFIX_NAME:
            if has_prefix_exclusive_lock() or has_prefix_lock():
                raise Exception("Cannot change prefix while a lock is acquired!")

        if prefix:
            self._prefix = Env.Prefix(prefix)
            self.PREFIX_NAME = prefix
            self._prefix.REPOS = get_repos()
        else:
            self.PREFIX_NAME = None
        return OLD_PREFIX_NAME

    def __init__(self):
        self.PREFIX_FILE = os.path.join(self.DATA_DIR, "prefix")
        self.REPOS_DIR = os.path.join(self.DATA_DIR, "repos")
        self.PYBUILD_CACHE_DIR = os.path.join(self.CACHE_DIR, "pybuild")
        self.DOWNLOAD_DIR = os.path.join(self.CACHE_DIR, "downloads")
        self.GLOBAL_PORTMOD_CONFIG = os.path.join(self.CONFIG_DIR, "portmod.conf")
        self.REPOS_FILE = os.path.join(self.CONFIG_DIR, "repos.cfg")
        self.REPOS = []

        tempfile.tempdir = None
        self.TMP_DIR = os.path.join(tempfile.gettempdir(), "portmod")
        self.PYBUILD_TMP_DIR = os.path.join(self.TMP_DIR, "pybuild")
        self.WARNINGS_DIR = os.path.join(self.PYBUILD_TMP_DIR, "messages")
        self.MESSAGES_DIR = os.path.join(self.PYBUILD_TMP_DIR, "warnings")
        self.TMP_VDB = os.path.join(self.TMP_DIR, "tmpdb")
        self.INDEX = os.path.join(self.CACHE_DIR, "index")

    def __setattr__(self, name: str, value: Any):
        if isinstance(value, str):
            os.environ["PORTMOD_" + name] = value
        super().__setattr__(name, value)


env = Env()

# Must be imported after env has been defined
from portmod.repos import get_local_repos  # noqa: E402


def refresh_env():
    global env

    env.__init__()  # type: ignore
    env.REPOS = list(get_local_repos().values())


env.REPOS = list(get_local_repos().values())



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
from logging import info, warning

from portmod.config import variable_data_dir
from portmod.config.sets import get_set
from portmod.globals import env
from portmod.loader import load_installed_pkg
from portmod.package import remove_pkg
from portmod.prefix import remove_prefix
from portmod.prompt import display_num_list, prompt_bool, prompt_num_multi
from portmodlib.fs import get_tree_size, onerror
from portmodlib.l10n import l10n


def destroy(args):
    assert env.PREFIX_NAME
    to_remove = []

    if args.remove_config and os.path.lexists(env.prefix().CONFIG_DIR):
        to_remove.append(env.prefix().CONFIG_DIR)

    if env.INTERACTIVE:
        if to_remove:
            print(l10n("destroy-directories"))
            display_num_list(
                to_remove,
                [l10n("size", size=get_tree_size(d) / 1024**2) for d in to_remove],
            )
            try:
                skip = prompt_num_multi(
                    l10n("destroy-exclude-prompt"),
                    len(to_remove),
                    cancel=True,
                )
            except EOFError:
                return

            for index in skip:
                del to_remove[index]
        else:
            print(l10n("no-directories-to-remove"))

        if not args.preserve_root:
            print(l10n("destroy-preserve-original", path=env.prefix().ROOT))

        print()
        if not prompt_bool(l10n("destroy-prompt", prefix=env.PREFIX_NAME)):
            return

    for directory in to_remove:
        print(l10n("removing-directory", path=directory))
        shutil.rmtree(directory, onerror=onerror)

    if not args.preserve_root:
        # Remove files installed by portmod
        # This only occurs for prefixes which were installed into an existing directory
        for atom in get_set("installed"):
            pkg = load_installed_pkg(atom)
            if pkg:
                remove_pkg(pkg)
            else:
                # This should not occur
                warning(f"Unable to load installed package {atom} during destruction")
        variable_data = variable_data_dir()
        info(l10n("removing-directory", path=variable_data))
        shutil.rmtree(variable_data, onerror=onerror)

    remove_prefix(env.PREFIX_NAME)


def add_destroy_parser(subparsers, parents):
    parser = subparsers.add_parser(
        "destroy", help=l10n("destroy-help"), parents=parents
    )
    parser.add_argument(
        "--preserve-root",
        help=l10n("destroy-preserve-root-help"),
        action="store_true",
    )
    parser.add_argument(
        "--remove-config",
        help=l10n("destroy-remove-config-help"),
        action="store_true",
    )
    parser.set_defaults(func=destroy)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import configparser
import csv
import getpass
import os
import re
from logging import info, warning
from stat import S_IWRITE
from typing import Mapping, NamedTuple, Optional

from portmod.functools import system_cache
from portmod.vdb import VDB, vdb_path
from portmodlib.fs import is_parent
from portmodlib.l10n import l10n

from .globals import env
from .repo.metadata import get_archs
from .repos import get_local_repos


class InvalidPrefix(RuntimeError):
    def __init__(self, prefix):
        self.prefix = prefix

    def __str__(self):
        return l10n("invalid-prefix", prefix=self.prefix)


class PrefixExistsError(RuntimeError):
    def __init__(self, prefix):
        self.prefix = prefix

    def __str__(self):
        return l10n("prefix-exists", prefix=self.prefix)


class Prefix(NamedTuple):
    arch: str
    path: Optional[str]


@system_cache
def get_prefixes() -> Mapping[str, Prefix]:
    """Returns a mapping of prefixes to their architectures"""
    prefixes = {}
    if os.path.exists(env.PREFIX_FILE):
        with open(env.PREFIX_FILE, "r", newline="") as file:
            # Ignore empty lines and comments
            lines = filter(
                bool, map(lambda x: re.sub("#.*", "", x).strip(), file.readlines())
            )
            csvdata = csv.reader(
                lines, delimiter=" ", dialect="unix", quoting=csv.QUOTE_MINIMAL
            )
            for row in csvdata:
                prefix = row[0]
                arch = row[1]
                directory = None
                if len(row) > 2:
                    directory = row[2]

                prefixes[prefix] = Prefix(arch, directory)

    return prefixes


def remove_prefix(prefix: str):
    """
    Removes a prefix the prefix file

    Does not modify the prefix's data
    """
    stat = os.stat(env.PREFIX_FILE)
    os.chmod(env.PREFIX_FILE, stat.st_mode | S_IWRITE, follow_symlinks=True)

    info(l10n("removing-prefix", prefix=prefix))
    with open(env.PREFIX_FILE) as file:
        lines = file.readlines()
        origlines = list(lines)
        for index, line in enumerate(lines):
            file_prefix, _, _ = line.partition(" ")
            if prefix == file_prefix:
                del lines[index]

    with open(env.PREFIX_FILE, "w") as file:
        if lines != origlines:
            file.writelines(lines)

    stat = os.stat(env.PREFIX_FILE)
    os.chmod(env.PREFIX_FILE, stat.st_mode - S_IWRITE, follow_symlinks=True)

    get_prefixes.cache_clear()


def add_prefix(prefix: str, arch: str, directory: Optional[str] = None):
    """
    Adds a new prefix

    args:
        prefix: Name of the prefix
        arch: Architecture of the prefix
        directory: Optional initial directory to store the prefix inside
                   If not included it will be stored with other portmod files
                   (~/.local/share/portmod on Linux)
                   Any files which are already in the directory will be backed up
                   if overwritten, and restored when the prefix is destroyed
    """

    invalid_prefixes = {"news", "sync", "mirror"}

    if prefix in invalid_prefixes:
        raise InvalidPrefix(prefix)

    if prefix in get_prefixes():
        raise PrefixExistsError(prefix)

    arch_options = set()
    for repo in get_local_repos().values():
        arch_options |= get_archs(repo.location)

    if arch not in arch_options:
        warning(l10n("unknown-arch", arch=arch))

    if directory:
        if not os.access(directory, os.W_OK):
            raise PermissionError(l10n("directory-not-writable", path=directory))

        existing_prefixes = get_prefixes()
        for name, existing_prefix in existing_prefixes.items():
            # directory cannot overlap with other prefixes
            existing_path = existing_prefix.path or os.path.join(env.DATA_DIR, name)
            if is_parent(directory, existing_path) or is_parent(
                existing_path, directory
            ):
                raise FileExistsError(
                    l10n(
                        "prefix-overlap",
                        newpath=directory,
                        oldpath=existing_path,
                        prefix=name,
                    )
                )

    # Add new prefix to the prefix file
    if os.path.exists(env.PREFIX_FILE):
        stat = os.stat(env.PREFIX_FILE)
        os.chmod(env.PREFIX_FILE, stat.st_mode | S_IWRITE, follow_symlinks=True)

    os.makedirs(os.path.dirname(env.PREFIX_FILE), exist_ok=True)

    with open(env.PREFIX_FILE, "a", newline="") as file:
        writer = csv.writer(
            file, delimiter=" ", dialect="unix", quoting=csv.QUOTE_MINIMAL
        )
        if directory is None:
            writer.writerow([prefix, arch])
        else:
            writer.writerow([prefix, arch, os.path.abspath(directory)])

    stat = os.stat(env.PREFIX_FILE)
    os.chmod(env.PREFIX_FILE, stat.st_mode - S_IWRITE, follow_symlinks=True)

    get_prefixes.cache_clear()

    env.set_prefix(prefix)

    # If the database already existed, it will not be recreated
    if os.path.exists(vdb_path()):
        print(l10n("database-exists", path=vdb_path()))
    else:
        import git

        with VDB() as gitrepo:
            # This repository is for local purposes only.
            # We don't want to worry about prompts for the user's gpg key password
            localconfig = gitrepo.config_writer()
            localconfig.set_value("commit", "gpgsign", False)
            USER = getpass.getuser()

            try:
                # May throw TypeError if GitPython<3.0.5
                globalconfig = git.config.GitConfigParser()
                globalconfig.get_value("user", "name")
                globalconfig.get_value("user", "email")
            except (TypeError, configparser.NoOptionError, configparser.NoSectionError):
                # Set the user name and email if they aren't in a global config
                localconfig.set_value("user", "name", f"{USER}")
                localconfig.set_value("user", "email", f"{USER}@example.com")

            localconfig.release()

    get_prefixes.cache_clear()
    info(l10n("initialized-prefix", prefix=prefix))



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import fnmatch
import os
import re
from typing import List, Pattern, Union


def pybuild_dumper(obj):
    # Serialize as best we can. Sets become lists and unknown objects are
    # just stringified
    if isinstance(obj, set):
        return list(sorted(obj))
    if hasattr(obj, "to_json"):
        return obj.to_json()
    return "{}".format(obj)


def fnmatch_list_to_re(patterns: List[str]) -> Pattern:
    def to_re(value: Union[str, Pattern]):
        """
        Converts fn-match string into a regular expression string

        Note that normpath may not work as expected with fn-match strings
        if forward-slashes are present inside bracketed ranges (e.g. [/../]).
        """
        if isinstance(value, Pattern):
            return value
        return fnmatch.translate(os.path.normpath(value))

    flags = 0
    if os.environ.get("CASE_INSENSITIVE_FILES", False):
        flags = re.IGNORECASE
    return re.compile("|".join(map(to_re, patterns)), flags=flags)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for directly loading pybuild files.
These functions should not be called directly.
See portmod.loader for functions to load pybuilds safely using a sandbox.
"""

import glob
import os
from copy import deepcopy
from logging import warning
from types import SimpleNamespace
from typing import Callable, Dict, Generator, Optional, cast

from portmod.config.mask import is_masked
from portmod.functools import install_cache, prefix_aware_cache, system_cache
from portmod.globals import env
from portmod.lock import vdb_lock
from portmod.parsers.manifest import Manifest
from portmod.repo import Repo, get_repo, get_repo_name, get_repo_root
from portmod.repo.metadata import get_categories, get_masters, is_pybuild_version_banned
from portmod.vdb import VDB, vdb_path
from portmodlib._loader import (
    SAFE_GLOBALS,
    __load_file_common,
    __load_installed_common,
    restricted_load,
)
from portmodlib.atom import Atom, FQAtom, QualifiedAtom, VAtom, Version, atom_sat
from portmodlib.l10n import l10n
from portmodlib.pybuild import FullPybuild

from .updates import get_moved


def commit_moved():
    """
    Moves installed packages which have been moved

    Should be run prior to entering the sandbox,
    as otherwise moved packages will not be visible
    """
    for repo in env.prefix().REPOS:
        moved = get_moved(repo)
        for source, dest in moved.items():
            atom = QualifiedAtom(source)

            path = os.path.join(vdb_path(), atom.C, atom.PN)
            if os.path.exists(path):
                destatom = QualifiedAtom(dest)
                destpath = os.path.join(vdb_path(), destatom.C, destatom.PN)
                for filepath in glob.glob(os.path.join(path, "*.pybuild")):
                    oldatom = get_atom_from_path(filepath)
                    if oldatom.R != repo.name:
                        with VDB(l10n("moved-package", src=source, dst=dest)) as vdb:
                            newfile = f"{destatom.PN}-{oldatom.PVR}.pybuild"
                            vdb.git.mv(filepath, os.path.join(path, newfile))
                            vdb.git.mv(path, destpath)
                            break


def find_installed_path(atom: Atom) -> Optional[str]:
    repo_path = vdb_path()

    # Handle renames:
    # Check all possible renames for this atom (all repos), and they are only valid if
    # the package comes from the repo matching the rename entry
    for repo in env.REPOS:
        moved = get_moved(repo)
        if atom.CPN in moved:
            moved_atom = QualifiedAtom(moved[atom.CPN])
            path = os.path.join(repo_path, moved_atom.C, moved_atom.PN)
            if os.path.exists(path):
                with open(os.path.join(path, "REPO"), "r") as file:
                    if repo.name == file.read().strip():
                        atom = moved_atom
                        break

    if atom.C:
        path = os.path.join(repo_path, atom.C, atom.PN)
        if os.path.exists(path):
            results = glob.glob(os.path.join(path, "*.pybuild"))
            assert len(results) == 1
            return results[0]
    else:
        for dirname in glob.glob(os.path.join(repo_path, "*")):
            path = os.path.join(repo_path, dirname, atom.PN)
            if os.path.exists(path):
                results = glob.glob(os.path.join(path, "*.pybuild"))
                assert len(results) == 1
                return results[0]
    return None


# Make sure this is only run once since _iterate_pybuilds is called many times
@system_cache
def check_repo_exists(name: str, location: str):
    if not os.path.exists(location):
        warning(
            l10n(
                "repo-does-not-exist-warning",
                name=name,
                path=location,
                command="portmod sync",
            )
        )


def _iterate_pybuilds(
    atom: Optional[Atom] = None,
    repo_name: Optional[str] = None,
    only_repo_root: Optional[str] = None,
) -> Generator[str, None, None]:
    """
    Iterates over pybuilds.

    If no repository is given, checks all available.
    If a repository name is given, checks that repository and its masters

    if only_repo_root is specified, only checks the repository at the given location
    """
    path = None
    repos = []
    if env.PREFIX_NAME:
        repos = env.prefix().REPOS
    else:
        repos = env.REPOS

    if repo_name is not None:
        repo = get_repo(repo_name)
        repos = [repo]
        for master in get_masters(repo.location):
            yield from _iterate_pybuilds(atom, master.name)
    elif only_repo_root:
        repos = [Repo(location=only_repo_root, name=get_repo_name(only_repo_root))]

    def valid_atom(desired_atom: Atom, other: Atom):
        if isinstance(desired_atom, FQAtom):
            return desired_atom == other
        else:
            return atom_sat(other, desired_atom)

    def try_move_atom(moved: Dict[str, str], atom: QualifiedAtom, repo):
        true_atom = atom
        if atom.CPN in moved:
            true_atom = QualifiedAtom(moved[atom.CPN])
            if atom.PVR:
                true_atom = QualifiedAtom(f"{true_atom}-{atom.PVR}")
        return true_atom

    for repo in repos:
        check_repo_exists(repo.name, repo.location)
        if atom:
            moved = get_moved(repo)
            if atom.C:
                true_atom = try_move_atom(moved, cast(QualifiedAtom, atom), repo)
                path = os.path.join(repo.location, true_atom.C, true_atom.PN)
                if path is not None:
                    for file in glob.iglob(os.path.join(path, "*.pybuild")):
                        if valid_atom(true_atom, get_atom_from_path(file)):
                            yield file
            else:
                for category in get_categories(repo.location):
                    true_atom = try_move_atom(moved, atom.with_category(category), repo)
                    path = os.path.join(repo.location, category, true_atom.PN)

                    if path is not None:
                        for file in glob.iglob(os.path.join(path, "*.pybuild")):
                            if valid_atom(true_atom, get_atom_from_path(file)):
                                yield file
        else:
            for file in glob.iglob(os.path.join(repo.location, "*", "*", "*.pybuild")):
                yield file


@system_cache
def _import_common(
    name: str,
    installed: bool,
    load_function: Callable,
    repo: Optional[str] = None,
) -> SimpleNamespace:
    """
    args:
        name: The import name as an absolute import path
        installed: Whether or not the package calling this is an installed package
        load_function: The function taking a file path and a keyword argument installed,
            indicating the installed status of the file to be loaded, to be used to load
            the common module

    returns:
        The Module as a SimpleNamespace
    """
    if len(name.split(".")) > 2:
        raise Exception(f"Invalid package {name}")
    _, module_name = name.split(".")
    base_atom = Atom(f"common/{module_name}")
    if installed:
        path = find_installed_path(base_atom)

    if installed and path:
        result = load_function(path, installed=True, repo=None)
    else:

        def get_version(file: str, allow_masked: bool) -> Optional[Version]:
            basename, _ = os.path.splitext(os.path.basename(file))
            atom = VAtom("common/" + basename)
            if not is_masked(atom, repo) or allow_masked:
                return atom.version
            return None

        files = list(_iterate_pybuilds(base_atom, repo_name=repo))

        max_version = None
        max_version_file = None

        def check_version(file: str, masked: bool):
            nonlocal max_version
            nonlocal max_version_file

            version = get_version(file, masked)
            if version is not None:
                if max_version is None or version > max_version:
                    max_version = version
                    max_version_file = file

        for file in files:
            check_version(file, False)

        if max_version is None:
            # If no versions were found, then try loading masked versions
            # Masked packages will still fail dependency resolution,
            for file in files:
                check_version(file, True)

        if not max_version:
            raise Exception(f"Could not find package {name}")

        result = load_function(max_version_file, installed=False, repo=repo)

    return SimpleNamespace(
        **{key: value for key, value in result.items() if not key.startswith("_")}
    )


def _safe_load_file(path: str, *, installed=False) -> FullPybuild:
    """
    Loads a pybuild file

    :param path: Path of the pybuild file
    """
    repo = get_repo_name(path) if not installed else None
    module = __safe_load_module(path, installed, repo)
    pkg = __load_file_common(path, module, installed)

    pkg.ATOM = get_atom_from_path(path)
    pkg.REPO = pkg.ATOM.R
    if installed:
        with vdb_lock():
            __load_installed_common(pkg, path)
            parent = os.path.dirname(path)
            pkg.INSTALLED_REBUILD_FILES = None
            path = os.path.join(parent, "REBUILD_FILES")
            if os.path.exists(path):
                try:
                    pkg.INSTALLED_REBUILD_FILES = Manifest(path)
                except ValueError:
                    warning(f"Failed to load {path}")
    return cast(FullPybuild, pkg)


@prefix_aware_cache
def pkg_exists(atom: Atom, *, repo_name: Optional[str] = None) -> bool:
    """Returns true if a package with the given atom can be found in the given repository"""
    return next(_iterate_pybuilds(atom, repo_name), None) is not None


def __safe_load_module(path: str, installed: bool, repo: Optional[str]):
    try:
        atom: Optional[FQAtom] = get_atom_from_path(path)
    except Exception:
        atom = None

    def get_info():
        assert atom
        return {
            "CATEGORY": atom.C,
            "P": atom.P,
            "PF": atom.PF,
            "PN": atom.PN,
            "PV": atom.PV,
            "PR": atom.PR,
            "PVR": atom.PVR,
            "__name__": "pybuild.info",
        }

    def get_pybuild():
        # File/InstallDir/Pybuild1 objects are instead passed as functions
        # to prevent modification of class attributes.
        # the instances are safe as long as underscored attributes are inaccessible

        from portmodlib._deprecated import File, InstallDir

        def file_func(*args, **kwargs):
            return File(*args, **kwargs)

        def install_dir_func(*args, **kwargs):
            return InstallDir(*args, **kwargs)

        # create simplified version of Pybuild to use for inheritance.
        # It has default values for all the pybuild fields.
        # We don't use FullPybuild directly to prevent its fields being modified
        # This class is generated within a function, and thus will not be the same
        # for any two packages.
        class Pybuild2(FullPybuild):
            """Generated Pybuild2 for use in non-sandboxed loader"""

            _PYBUILD_VER = 2

        class Pybuild1(FullPybuild):
            """Generated Pybuild2 for use in non-sandboxed loader"""

            _PYBUILD_VER = 1
            DATA_OVERRIDES = ""
            INSTALL_DIRS = []
            REBUILD_FILES = []
            TIER = "a"

            def __init__(self):
                super().__init__()
                if type(self.TIER) is int:
                    self.TIER = str(self.TIER)

        imports = {
            "File": file_func,
            "InstallDir": install_dir_func,
            "DOWNLOAD_DIR": env.DOWNLOAD_DIR,
            "version_gt": None,
            "find_file": None,
            "use_reduce": None,
            "check_required_use": None,
            "patch_dir": None,
            "get_masters": None,
            "apply_patch": None,
            "__name__": "pybuild",
        }
        pybuild_versions = {
            1: Pybuild1,
            2: Pybuild2,
        }

        repo_root = get_repo_root(path)
        for version, class_ in pybuild_versions.items():
            # If the package is an installed package, or the version is not banned
            if not repo_root or not is_pybuild_version_banned(repo_root, version):
                imports[f"Pybuild{version}"] = class_

        return imports

    def get_typing():
        import typing

        result = {}

        for attr, value in typing.__dict__.items():
            if not attr.startswith("_") and not isinstance(value, type(typing)):
                result[attr] = deepcopy(value)

        return result

    def _import(name, _globs, _loc, _fromlist, _level):
        if name.startswith("common."):
            return deepcopy(_import_common(name, installed, __safe_load_module, repo))
        if name == "common":
            raise RuntimeError(
                "Cannot import from common directly. Import from a submodule instead"
            )
        if name == "pybuild":
            return SimpleNamespace(**get_pybuild())
        if name == "pybuild.info":
            return SimpleNamespace(**get_info())
        if name == "typing":
            return SimpleNamespace(**get_typing())

        class NoAttr(type):
            def __init__(cls, typename, bases, dictionary):
                super().__init__(typename, bases, dictionary)
                # Note: trailing space should mean that this error can't be ignored by subclassing this
                # with a class named NoAttr. This probably isn't a security issue.
                if cls.__name__ != "NoAttr ":
                    raise TypeError(
                        f"{typename} cannot inherit from {cls.item}. "
                        "Only classes imported from `common.*` and `pybuild` can be inherited from"
                    )

            def __getattr__(cls, attr):
                # Necessary so that the object can be copied
                # Note that RestrictedPython prevents use of such attributes through ast parsing
                if attr.startswith("_"):
                    return super().__getattr__(attr)  # type: ignore  # pylint: disable=no-member
                raise RuntimeError(
                    f"Cannot use {name}.{attr}: Complex code in the global scope is not allowed!"
                )

        if _fromlist:
            return SimpleNamespace(
                **{
                    item: NoAttr("NoAttr ", (object,), {"item": item})
                    for item in _fromlist
                }
            )
        else:
            return NoAttr("NoAttr ", (object,), {})

    filename, _ = os.path.splitext(os.path.basename(path))

    with open(path, "r", encoding="utf-8") as file:
        code = file.read()
        tmp_globals = deepcopy(SAFE_GLOBALS)
        tmp_globals["__builtins__"]["__import__"] = _import
        tmp_globals["__name__"] = filename
        restricted_load(code, path, tmp_globals)
        for _global in tmp_globals.values():
            # Note: we can't use hasattr here, as we want to set the attribute for this classmethod
            if isinstance(_global, type) and "__file__" not in _global.__dict__:
                try:
                    setattr(_global, "__file__", path)
                except:  # noqa
                    ...

    return tmp_globals


def _iterate_installed() -> Generator[str, None, None]:
    for path in glob.glob(os.path.join(vdb_path(), "*", "*", "*.pybuild")):
        yield path


@install_cache
def get_atom_from_path(path: str) -> FQAtom:
    repopath, filename = os.path.split(os.path.abspath(os.path.normpath(path)))
    atom, _ = os.path.splitext(filename)
    repopath, _ = os.path.split(repopath)
    repopath, C = os.path.split(repopath)
    try:
        repo_name = get_repo_name(repopath)
    except FileNotFoundError as e:
        path = os.path.join(os.path.dirname(path), "REPO")
        if os.path.exists(path):
            with open(path, "r") as file:
                repo_name = file.read().strip() + "::installed"
        else:
            raise e
    return FQAtom(f"{C}/{atom}::{repo_name}")



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3


import os
from logging import info

from portmod.config import get_config, set_config_value
from portmod.config.profiles import get_profile_path, profile_exists, set_profile
from portmod.globals import env
from portmod.prefix import add_prefix
from portmod.prompt import prompt_bool, prompt_num, prompt_num_multi
from portmod.repo import get_repo
from portmod.repo.metadata import get_archs
from portmod.repos import add_repo
from portmod.sync import sync
from portmodlib.colour import bright
from portmodlib.l10n import l10n

from .select import add_prefix_repo, list_profiles, list_repos


def init(args):
    """Initializes a prefix"""
    add_prefix(args.prefix, args.arch, args.directory)

    if env.INTERACTIVE:
        # If prefix previously existed and the configuration was preserved from when it
        # was removed, configuration could exist already.
        # If so, display configuration and prompt user to keep it or go through the
        # normal selection process
        if profile_exists():
            print(l10n("existing-configuration", prefix=args.prefix))
            enabled = " ".join(get_config()["REPOS"])
            print()
            print(f'    REPOS = "{enabled}"')
            print(f'    profile = "{get_profile_path()}"')
            print()
            if prompt_bool(l10n("existing-configuration-prompt")):
                return

            # Ensure that existing repositories get cleared so that only the new ones
            # will be included
            set_config_value("REPOS", "")

        info("")
        info(l10n("init-preamble"))
        info("")

        repos = list_repos(args.arch)
        if len(repos) == 1:
            info("")
            info(bright(l10n("init-single-repo", arch=args.arch, repo=repos[0].name)))
            info("")
            selected = [0]
        else:
            selected = prompt_num_multi(l10n("init-repositories-prompt"), len(repos))

        for index in selected:
            info(f"portmod {args.prefix} select repo add {repos[index].name}")
            add_prefix_repo(repos[index].name)
            add_repo(repos[index])

        # Re-initialize prefix with the new repositories
        env.set_prefix(args.prefix)

        if selected:
            info("portmod sync " + " ".join([repos[x].name for x in selected]))
            sync([get_repo(repos[x].name) for x in selected])

        print()
        profiles = list_profiles(args.arch)
        index = prompt_num(l10n("init-profile-prompt"), len(profiles))
        info(f"portmod {args.prefix} select profile set {index}")
        set_profile(profiles[index])

        info("")
        info(l10n("init-subcommands"))
        info(f"    portmod {args.prefix} select profile")
        info(f"    portmod {args.prefix} select repo")
    else:
        if not profile_exists():
            info(l10n("init-non-interactive-postamble"))
            info(f"    portmod {args.prefix} select profile")
            info(f"    portmod {args.prefix} select repo")


def add_init_parser(subparsers, parents):
    parser = subparsers.add_parser("init", help=l10n("init-help"), parents=parents)
    parser.add_argument(
        "prefix", metavar=l10n("prefix-placeholder"), help=l10n("init-prefix-help")
    )
    try:
        meta_repo = get_repo("meta")
        if not os.path.exists(meta_repo.location) and not env.TESTING:
            sync([meta_repo])
            if not env.REPOS:
                env.REPOS = [meta_repo]
        parser.add_argument(
            "arch",
            help=l10n("init-arch-help"),
            choices={arch for repo in env.REPOS for arch in get_archs(repo.location)}
            or None,
        )
    except Exception:
        parser.add_argument("arch", help=l10n("init-arch-help"))

    parser.add_argument(
        "directory",
        metavar=l10n("directory-placeholder"),
        help=l10n(
            "init-directory-help",
            local=env.DATA_DIR.replace(os.path.expanduser("~"), "~"),
        ),
        nargs="?",
    )
    parser.set_defaults(func=init)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
from contextlib import ContextDecorator
from typing import Optional

from .config import variable_data_dir
from .lock import vdb_lock


def vdb_path() -> str:
    """
    Returns the VDB path for the current prefix
    """
    return os.path.join(variable_data_dir(), "db")


class VDB(ContextDecorator):
    def __init__(self, commit_message: Optional[str] = None):

        self.lock = vdb_lock(write=True)
        self.message = commit_message

    def __enter__(self):
        # Slow import
        import git

        self.lock.__enter__()
        # Don't set the gitrepo until we have the write lock
        self.gitrepo = git.Repo.init(vdb_path())
        return self.gitrepo

    def __exit__(self, *_exc):
        if self.message is not None and (
            not self.gitrepo.heads or self.gitrepo.git.diff("HEAD", cached=True)
        ):
            self.gitrepo.git.commit(m=self.message)
        self.lock.__exit__()
        return False



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from datetime import datetime

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from portmod.globals import get_authors, get_version  # noqa

year = datetime.now().year


# -- Project information -----------------------------------------------------

project = "Portmod"
copyright = f"2019-{year}, Portmod Authors"
authors = get_authors()
author = ", ".join(authors)

# The short X.Y version
version = f"Portmod {get_version()}"

# The full version, including alpha/beta/rc tags
release = f"Portmod {get_version()}"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinx_rtd_theme",
    "sphinxarg.ext",
    "sphinxcontrib.apidoc",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "navigation_depth": -1,
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# These paths are either relative to html_static_path
# or fully qualified paths (eg. https://...)
html_css_files = [
    "css/custom.css",
]


# -- Extension configuration -------------------------------------------------

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
}

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
always_document_param_types = True

# API docs

apidoc_module_dir = "../pybuild"
apidoc_output_dir = "dev/pybuild"
apidoc_separate_modules = True
apidoc_toc_file = False



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the setup systems
"""

import os
import sys
from contextlib import redirect_stdout
from io import StringIO

import pytest

from portmod._cli.main import main
from portmod.config.profiles import get_profile_path
from portmod.globals import env
from portmod.prefix import get_prefixes
from portmod.repos import get_local_repos

from .env import setup_env, setup_no_prefix, tear_down_env


@pytest.fixture(autouse=False)
def setup():
    """
    Sets up test repo for querying
    """
    yield setup_no_prefix()
    tear_down_env()


@pytest.fixture(autouse=False)
def setup_prefix():
    """
    Sets up test repo for querying
    """
    yield setup_env("test")
    tear_down_env()


def test_init_prefix(setup):
    """Tests prefix creation"""
    sys.argv = ["portmod", "init", "test2", "test", "--no-confirm"]
    main()

    assert get_prefixes()["test2"].arch == "test"
    sys.argv = ["portmod", "test2", "destroy", "--no-confirm"]
    main()
    assert "test2" not in get_prefixes()


def test_init_prefix_interactive(setup):
    """Tests prefix creation"""
    env.set_prefix(None)
    sys.argv = ["portmod", "init", "test2", "test"]
    oldstdin = sys.stdin
    # Note: these are hardcoded indices which may change
    # 3 is for the test repository, 0 for the first profile in the list
    # (doesn't matter which one, we're just testing that the profile is selected correctly)

    # The test repository has two copies of the test-extra repository which gets pulled in
    # as repos 1 and 2 to test for a bug where running sync with a prefix set will cause an
    # error when a new repository gets synched twice, as in this case where test-extra and
    # text-extra2 both depend on blank.
    stringio = StringIO("1,2,3\n0\n")
    env.INTERACTIVE = True
    sys.stdin = stringio
    main()
    sys.stdin = oldstdin
    env.INTERACTIVE = False

    env.set_prefix("test2")
    assert os.path.exists(get_profile_path())

    assert get_prefixes()["test2"].arch == "test"
    sys.argv = ["portmod", "test2", "destroy", "--no-confirm"]
    main()
    assert "test2" not in get_prefixes()


def test_add_repo(setup_prefix):
    """Tests adding repositories automatically"""
    stringio = StringIO()
    with redirect_stdout(stringio):
        sys.argv = ["portmod", "test", "select", "repo", "list"]
        main()
    assert "blank" in stringio.getvalue()

    sys.argv = ["portmod", "test", "select", "repo", "add", "blank"]
    main()

    assert any(repo.name == "blank" for repo in env.REPOS)
    assert "blank" in get_local_repos()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Functions for interacting with the OpenMW VFS
"""

import os
import shutil
from functools import wraps
from tempfile import gettempdir
from typing import Any, Callable, Dict, List, TypeVar, cast

from portmodlib.archives import extract_archive_file, list_archive
from portmodlib.fs import ci_exists, is_parent
from portmodlib.functools import prefix_aware_cache
from portmodlib.globals import root
from portmodlib.l10n import l10n
from portmodlib.parsers.list import read_list

F = TypeVar("F", bound=Callable[..., Any])
_VFS_CACHE_FUNCS = []


def vfs_cache(func: F) -> F:
    """
    A variant of functools.lru_cache which treats the prefix as if it were an argument
    """
    global _VFS_CACHE_FUNCS

    @wraps(func)
    @prefix_aware_cache
    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    _VFS_CACHE_FUNCS.append(inner)

    return cast(F, inner)


def clear_vfs_cache():
    for func in _VFS_CACHE_FUNCS:
        func.cache_clear()  # type: ignore


@vfs_cache
def find_file(name: str) -> str:
    """
    Locates the path of a file within the OpenMW virtual file system

    .. warning::
        This function has been deprecated as of Portmod 2.4.
        It will be removed in Portmod 3.0

    args:
        name: The relative path within the VFS to search for

    returns:
        The absolute path of the file
    """
    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting
    if os.environ.get("VFS"):
        for directory in reversed(get_vfs_dirs()):
            path = ci_exists(os.path.join(directory, name))
            if path:
                return path

        for archive in reversed(get_vfs_archives()):
            contents = list_archive(archive)
            for file in contents:
                if os.path.normpath(file).lower() == os.path.normpath(name).lower():
                    return extract_archive_file_to_tmp(archive, file)
    else:
        path = ci_exists(name, prefix=root())
        if path:
            return path

    raise FileNotFoundError(name)


@vfs_cache
def list_dir(name: str) -> List[str]:
    """
    Locates all path of files matching the given pattern within the OpenMW
    virtual file system

    .. warning::
        This function has been deprecated as of Portmod 2.4.
        It will be removed in Portmod 3.0

    args:
        name: The relative path of the directory within the VFS
    returns:
        A list of files contained within the directory
    """
    files: Dict[str, str] = {}
    normalized = os.path.normpath(name).lower()

    # FIXME: This should respect the CASE_INSENSITIVE_FILES setting

    if os.environ.get("VFS"):
        for directory in reversed(get_vfs_dirs()):
            path = ci_exists(os.path.join(directory, normalized))
            if path:
                for file in os.listdir(path):
                    if file.lower() not in files:
                        files[file.lower()] = file

        for archive in reversed(get_vfs_archives()):
            contents = list_archive(archive)
            for file in contents:
                if is_parent(os.path.normpath(file).lower(), normalized):
                    suffix = os.path.relpath(os.path.normpath(file).lower(), normalized)
                    component, _, _ = suffix.partition(os.sep)
                    files[component] = component
    else:
        path = ci_exists(normalized, prefix=root())
        if path:
            for file in os.listdir(path):
                if file.lower() not in files:
                    files[file.lower()] = file

    return sorted(files.values())


def _cleanup_tmp_archive_dir():
    path = os.path.join(gettempdir(), ".archive_files")
    if os.path.exists(path):
        shutil.rmtree(path)


def extract_archive_file_to_tmp(archive: str, file: str) -> str:
    """Extracts the given file from the archive and places it in a temprorary directory"""
    temp = gettempdir()
    output_dir = os.path.join(
        temp, ".archive_files", os.path.basename(archive), os.path.dirname(file)
    )
    os.makedirs(output_dir, exist_ok=True)
    result_file = os.path.join(output_dir, os.path.basename(file))
    extract_archive_file(archive, file, output_dir)
    if not os.path.exists(result_file):
        raise Exception(l10n("archive-extraction-failed", file=file, dest=result_file))
    return result_file


@vfs_cache
def get_vfs_dirs() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    dirs: List[str] = read_list(
        os.path.join(root(), os.environ["VARIABLE_DATA"], "vfs")
    )
    return dirs


@vfs_cache
def get_vfs_archives() -> List[str]:
    """Returns an ordered list of the VFS directories, in reverse order of priority"""
    archives: List[str] = read_list(
        os.path.join(root(), os.environ["VARIABLE_DATA"], "vfs-archives")
    )
    return archives



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3
"""
Module building on behaviour from the builtin functools  module
"""

from functools import lru_cache, wraps
from typing import Callable, Generic, Hashable, Optional, TypeVar, cast

from portmodlib._deprecated.vfs import clear_vfs_cache

from .globals import env

_T = TypeVar("_T")

_SYSTEM_CACHE_FUNCS = []
_INSTALL_CACHE_FUNCS = []


class _lru_cache_wrapper(Generic[_T]):
    __wrapped__: Callable[..., _T]

    def __call__(self, *args: Hashable, **kwargs: Hashable) -> _T:  # type: ignore
        ...

    def cache_clear(self) -> None:
        ...


def system_cache(func: Callable[..., _T]) -> _lru_cache_wrapper[_T]:
    """
    A variant of lru_cache which gets registered so that caches can be cleared more easily

    This should be used for cached functions which are dependent on the host system, and
    may need to be invalidated in long-running processses.
    """
    global _SYSTEM_CACHE_FUNCS

    @wraps(func)
    @lru_cache()
    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    _SYSTEM_CACHE_FUNCS.append(inner)

    return cast(_lru_cache_wrapper[_T], inner)


def prefix_aware_cache(func: Callable[..., _T]) -> _lru_cache_wrapper[_T]:
    """
    A variant of functools.lru_cache which treats the prefix as if it were an argument

    Like system_cache, but for data which is specific to a certain prefix
    """
    global _SYSTEM_CACHE_FUNCS

    @wraps(func)
    @lru_cache(maxsize=None)
    def inner(_prefix: Optional[str], *args, **kwargs):
        return func(*args, **kwargs)

    @wraps(func)
    def prefix_wrapper(*args, **kwargs):
        return inner(env.PREFIX_NAME, *args, **kwargs)

    prefix_wrapper.cache_clear = inner.cache_clear  # type: ignore
    _SYSTEM_CACHE_FUNCS.append(prefix_wrapper)  # type: ignore
    return cast(_lru_cache_wrapper[_T], prefix_wrapper)


def install_cache(func: Callable[..., _T]) -> _lru_cache_wrapper[_T]:
    """
    A variant of functools.lru_cache which treats the prefix as if it were an argument

    like prefix_aware_cache, but for data which portmod itself may change when installing
    packages.
    """
    global _INSTALL_CACHE_FUNCS

    @wraps(func)
    @prefix_aware_cache
    def inner(*args, **kwargs):
        return func(*args, **kwargs)

    _INSTALL_CACHE_FUNCS.append(inner)

    return cast(_lru_cache_wrapper[_T], inner)


def clear_install_cache():
    for func in _INSTALL_CACHE_FUNCS:
        func.cache_clear()
    clear_vfs_cache()


def clear_system_cache():
    for func in _SYSTEM_CACHE_FUNCS:
        func.cache_clear()
    clear_vfs_cache()



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Download and source related tests
"""

from portmod.pybuild import parse_arrow


def test_simple_source_parsing():
    """
    Simple source test of basic functionality
    """
    path = "http://simplesource.com/foo.zip"
    results = parse_arrow(path.split())

    assert len(results) == 1
    assert results[0].url == path
    assert results[0].name == "foo.zip"
    assert results[0].basename == "foo"


def test_arrow_source_parsing():
    """
    Tests parsing arrow notation
    """
    path = "http://simplesource.com/foo.zip"
    path2 = "http://othersource.com/foo.zip"
    results = parse_arrow(f"{path} -> bar.zip {path2}".split())

    assert len(results) == 2
    assert results[0].url == path
    assert results[0].name == "bar.zip"
    assert results[0].basename == "bar"
    assert results[1].url == path2
    assert results[1].name == "foo.zip"
    assert results[1].basename == "foo"



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3
"""
Module that interacts with the python-based portmod.conf config file
"""

import os
from copy import deepcopy
from textwrap import fill
from typing import Any, Dict, Optional

from portmod.globals import env, get_version
from portmod.parsers.flags import collapse_flags
from portmodlib._loader import MINIMAL_GLOBALS, Policy
from portmodlib.l10n import l10n

__COLLAPSE_KEYS = {
    "USE",
    "ACCEPT_KEYWORDS",
    "INFO_VARS",
    "INFO_PACKAGES",
    "USE_EXPAND",
    "USE_EXPAND_HIDDEN",
    "PROFILE_ONLY_VARIABLES",
    "CACHE_FIELDS",
    "REPOS",
}
__OVERRIDE_KEYS = {
    "ACCEPT_LICENSE",
    "TEXTURE_SIZE",
    "PORTMOD_MIRRORS",
    "CASE_INSENSITIVE_FILES",
    "OMWMERGE_DEFAULT_OPTS",
    "INSTALL_DEST",
    "DOC_DEST",
    "VARIABLE_DATA",
    "VFS",
}

__USER_ONLY_VARIABLES = {"TMP", "TMPDIR", "TEMP", "_SANDBOX_PERMISSIONS", "MERGE_TOOL"}


def _comment_wrap(text: str) -> str:
    return "\n".join(
        [
            fill(paragraph, width=80, initial_indent="# ", subsequent_indent="# ")
            # Squelch text together as one paragraph, except for double linebreaks, which define
            # new paragraphs
            for paragraph in text.split("\n\n")
        ]
    )


def _create_config_placeholder():
    """
    Creates a placeholder config file to help the user initialize their
    config file for the first time
    """
    os.makedirs(env.prefix().CONFIG_DIR, exist_ok=True)
    version = get_version()
    with open(env.prefix().CONFIG, "w") as file:
        config_string = (
            # FIXME: Depending on how the wiki is localized, we may need to have wiki_page included in the localization
            # For now it is excluded so that it can be changed more easily across localizations
            _comment_wrap(
                l10n(
                    "config-placeholder-header",
                    version=version,
                    info_command=f"`portmod {env.PREFIX_NAME} info`",
                    wiki_page="https://gitlab.com/portmod/portmod/wikis/Portmod-Config",
                )
            )
            + "\n\n"
            + _comment_wrap(l10n("config-placeholder-global-use"))
            + '\n# USE=""\n\n'
            + _comment_wrap(l10n("config-placeholder-texture-size", default="min"))
            + '\n# TEXTURE_SIZE="min"\n\n'
            + _comment_wrap(l10n("config-placeholder-accept-keywords"))
            + '\n# ACCEPT_KEYWORDS="openmw"\n\n'
            + _comment_wrap(l10n("config-placeholder-accept-license"))
            + '\n# ACCEPT_LICENSE="* -EULA"\n\n'
            + _comment_wrap(l10n("config-placeholder-openmw-config"))
            + "\n\n"
            + _comment_wrap(l10n("config-placeholder-morrowind-path"))
        )
        print(config_string, file=file)


def read_config(path: str, old_config: Dict[str, Any], *, user: bool = False) -> Dict:
    """
    Reads a config file and converts the relevant fields into a dictionary
    """
    # Slow import
    from RestrictedPython import compile_restricted

    if user:
        # User-config should auto-detect encoding such that it matches the
        # user's editor
        encoding = None
    else:
        # Repository files should always be utf-8
        encoding = "utf-8"

    with open(path, "r", encoding=encoding) as file:
        config = file.read()

    byte_code = compile_restricted(config, filename=path, mode="exec", policy=Policy)

    glob = deepcopy(MINIMAL_GLOBALS)
    glob["__builtins__"]["join"] = os.path.join
    new_config = old_config.copy()
    try:
        exec(byte_code, glob, new_config)
    except NameError as e:
        print(l10n("exec-error", error=e, file=path))
    except SyntaxError as e:
        print(l10n("exec-error", error=e, file=path))

    merged = old_config.copy()

    def line_of_key(key: str) -> Optional[int]:
        for index, line in enumerate(config.split("\n")):
            if key in line:
                return index
        return None

    def check_key(key):
        nonlocal new_config, old_config, merged
        if (
            user
            and key in merged.get("PROFILE_ONLY_VARIABLES", [])
            and new_config.get(key) is not None
            and new_config.get(key) != old_config.get(key)
        ):
            raise UserWarning(
                f"{path}:{line_of_key(key)}\n" + l10n("reserved-variable", key=key)
            )

        if (
            not user
            and key in __USER_ONLY_VARIABLES
            and new_config.get(key) is not None
            and new_config.get(key) != old_config.get(key)
        ):
            raise UserWarning(
                f"{path}:{line_of_key(key)}\n" + l10n("user-reserved-variable", key=key)
            )

    for key in __COLLAPSE_KEYS:
        check_key(key)

        if isinstance(new_config.get(key, ""), str):
            new_config[key] = set(new_config.get(key, "").split())
        merged[key] = collapse_flags(merged.get(key, set()), new_config.get(key, set()))

    for key in __OVERRIDE_KEYS:
        check_key(key)

        if key in new_config and new_config[key]:
            merged[key] = new_config.get(key)

    for key in old_config.keys() | new_config.keys():
        check_key(key)

        if key not in __COLLAPSE_KEYS | __OVERRIDE_KEYS:
            if key in new_config:
                merged[key] = new_config[key]
            else:
                merged[key] = old_config[key]

    return merged



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""A module for parsing package updates"""

import os
from collections import defaultdict
from typing import Any, Dict, Optional

from portmod.globals import env
from portmod.repos import Repo
from portmodlib.parsers.list import read_list

_UPDATES: Optional[Dict[str, Dict[str, Any]]] = None


def get_moved(repo: Repo) -> Dict[str, str]:
    global _UPDATES
    if _UPDATES is None:
        _UPDATES = {}
        for repo in env.REPOS:
            parse_updates(repo)

    assert _UPDATES is not None
    return _UPDATES.get(repo.name, {}).get("move", {})


def parse_updates(repo: Repo):
    global _UPDATES
    path = os.path.join(repo.location, "profiles", "updates")
    repo_updates: Dict[str, Any] = defaultdict(dict)

    def parse_file(filename: str):
        result: Dict[str, Dict[str, Any]] = defaultdict(dict)
        for line in read_list(filename):
            command, _, arguments = line.partition(" ")
            if command == "move":
                source, target = arguments.split(" ")
                source = source.strip()
                target = target.strip()
                result["move"][source] = target
        return result

    if os.path.exists(path):
        for filename in os.listdir(path):
            for key, value in parse_file(os.path.join(path, filename)).items():
                repo_updates[key].update(value)

    assert _UPDATES is not None
    _UPDATES[repo.name] = repo_updates



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module with localization helpers"""

import locale
import os
from functools import lru_cache
from typing import List

from portmodlib.portmod import l10n_lookup

# Only used in debug builds of the rust library
_DEBUG_L10N_DIR = os.path.join(os.path.dirname(os.path.dirname(__file__)), "l10n")


@lru_cache()
def _default_locale():
    """Returns the default locale"""
    try:
        locale.setlocale(locale.LC_ALL, "")
        lang, _ = locale.getdefaultlocale()
        if lang is None:
            return "en_GB"
        return str(lang)
    except locale.Error:
        # If default fails, fall back to the C locale, which should hopefully work
        locale.setlocale(locale.LC_ALL, "C")
        return "en_GB"


def l10n(msg_id: str, **kwargs) -> str:
    """
    Fetches a localised message and formats it with the given arguments

    Wrapper around portmodlib.portmod.l10n_lookup, in the Rust extension,
    which itself is built on top of
    `fluent_templates <https://github.com/XAMPPRocky/fluent-templates>`__

    args:
        msg_id: The message identifier
        kwargs: Arguments passed to the fluent formatter
    """
    # Get locale before formatting numbers so that LC_ALL gets set properly
    default = _default_locale()

    # TODO: Replace this with fluent formatting when fluent-rs better supports floats
    for key, value in kwargs.items():
        if isinstance(value, float):
            kwargs[key] = f"{value:n}"

    result = l10n_lookup(default, msg_id, kwargs)
    if result:
        return result
    raise RuntimeError(f"No Localization exists for id {msg_id}")


@lru_cache()
def get_locales(separator: str = "-") -> List[str]:
    """Returns detected locales in the form suitable for the repository"""
    locales = []
    parts = _default_locale().replace("-", "_").split("_")
    if len(parts) == 2:
        locales.append(parts[0] + separator + parts[1])
    locales.append(parts[0])
    # Default (lowest priority) is en
    locales.append("en")
    return locales



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from portmod.globals import env
from portmod.lock import exclusive_lock
from portmod.merge import configure
from portmod.prompt import prompt_num_multi
from portmod.query import display_search_results, query
from portmodlib.colour import bright, lblue
from portmodlib.l10n import l10n


def search_main(args):
    pkgs = list(
        query(
            " ".join(args.query),
        )
    )
    display_search_results(pkgs, numbers=True)

    if env.INTERACTIVE:
        try:
            selection = selection_prompt([pkg.cpn for pkg in pkgs])
        except EOFError:
            return
        if selection:
            with exclusive_lock():
                configure(selection)


def selection_prompt(atoms):
    # Allow selecting numbers between 0 and the number of mods
    selection = prompt_num_multi(
        bright(lblue(":: ")) + bright("Packages to install (eg: 1,2,3)"), len(atoms)
    )
    # Iterate through the selected numbers, and get
    # the atoms that correspond to those numbers
    mod_selection = [atoms[number - 1] for number in selection]
    return [index for index in mod_selection]


def add_search_parser(subparsers, parents):
    parser = subparsers.add_parser("search", help=l10n("search-help"), parents=parents)

    parser.add_argument(
        "query",
        nargs="+",
        metavar=l10n("query-placeholder"),
        help=l10n("search-query-help"),
    )
    parser.set_defaults(func=search_main)



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Helper functions for interacting with package licenses
"""

import os
from typing import AbstractSet, List, Union

from portmod.config import get_config
from portmod.config.use import get_use
from portmod.functools import system_cache
from portmod.globals import env
from portmod.parsers.flags import get_flags
from portmod.pybuild import Pybuild
from portmod.repo import get_repo
from portmod.repo.metadata import get_license_groups
from portmodlib.usestr import use_reduce


def _get_pkg_license_groups(pkg: Pybuild):
    if pkg.INSTALLED:
        return get_license_groups(get_repo(pkg.REPO).location)
    else:
        if not pkg.REPO_PATH:
            raise RuntimeError(
                "Checking license of a package with no repository! "
                "This should not happen!"
            )
        return get_license_groups(pkg.REPO_PATH)


def _user_package_accept_license_path() -> str:
    return os.path.join(env.prefix().CONFIG_DIR, "package.accept_license")


def is_license_accepted(pkg: Pybuild, enabled_flags: AbstractSet[str]) -> bool:
    """
    Returns true if the package's license(s) are accepted by the user's configuration

    For a license to be accepted, it must be both listed, either explicitly,
    part of a group, or with the * wildcard  and it must not be blacklisted
    by a license or license group prefixed by a '-'
    """
    license_groups = _get_pkg_license_groups(pkg)

    ACCEPT_LICENSE = get_config()["ACCEPT_LICENSE"].split()
    package_accept_license = get_flags(_user_package_accept_license_path(), pkg.ATOM)

    def accepted(group: Union[str, List]) -> bool:
        if not group:
            return True

        if isinstance(group, str):
            allowed = False
            # Check if license is allowed by anything in ACCEPT_LICENSE
            for license in list(ACCEPT_LICENSE) + list(package_accept_license):
                if license.startswith("-") and (
                    license == group
                    or (license[1] == "@" and group in license_groups[license[2:]])
                ):
                    # not allowed if matched by this, but can be overridden later
                    allowed = False
                if license == "*":
                    allowed = True
                if license.startswith("@") and group in license_groups[license[1:]]:
                    allowed = True
            return allowed
        if group[0] == "||":
            return any(accepted(license) for license in group)

        return all(accepted(license) for license in group)

    return accepted(use_reduce(pkg.LICENSE, enabled_flags, opconvert=True))


@system_cache
def has_eula(package: Pybuild) -> bool:
    groups = _get_pkg_license_groups(package)

    # FIXME: This should be reworked.
    # For one thing, this doesn't currently handle || operators
    return any(
        license_name in groups.get("EULA", set())
        for license_name in use_reduce(
            package.LICENSE, get_use(package)[0], get_use(package)[1], flat=True
        )
    )



# Copyright 2022 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests file master detection
"""

import pytest

from portmod.loader import load_pkg
from portmod.package import install_pkg
from portmodlib.atom import Atom

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup_repo():
    """sets up and tears down test environment"""
    yield setup_env("test")
    tear_down_env()


def test_unpack():
    """
    Tests that src_unpack works as expected for Pybuild2

    Checks are inside the package
    """
    install_pkg(load_pkg(Atom("=test/quill-of-feyfolken-2.0.2-r1"))[0], set())



# Copyright 2019-2021 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for interacting with the user's profile
and iterating over the various associated directories
"""

import os
from typing import List, Set, Tuple

from portmod.functools import prefix_aware_cache
from portmod.globals import env
from portmod.repo import get_repo_root
from portmodlib.atom import Atom
from portmodlib.parsers.list import read_list


def profile_link() -> str:
    return os.path.join(env.prefix().CONFIG_DIR, "profile")


def set_profile(path: str) -> None:
    os.makedirs(env.prefix().CONFIG_DIR, exist_ok=True)
    linkpath = profile_link()
    if os.path.lexists(linkpath):
        os.unlink(linkpath)
    os.symlink(path, linkpath)


def profile_exists() -> bool:
    return bool(env.PREFIX_NAME and os.path.exists(profile_link()))


def get_profile_path() -> str:
    """Returns the path to the profile directory"""
    profilepath = profile_link()
    if not os.path.exists(profilepath) or not os.path.islink(profilepath):
        raise FileNotFoundError(
            f"{profilepath} does not exist.\n"
            "Please choose a profile before attempting to install packages"
        )

    return os.path.realpath(os.readlink(profilepath))


def _profile_relative_path() -> Tuple[str, str]:
    """
    Splits the profile path into the profile base
    and a relative path within the profile directory
    """
    fullpath = get_profile_path()
    repo_root = get_repo_root(fullpath)
    if not repo_root:
        raise RuntimeError(
            "Profile symlinks must point to a path inside a repository profile directory!"
        )
    profile_root = os.path.join(repo_root, "profiles")
    return profile_root, os.path.relpath(fullpath, start=profile_root)


def get_profile_name() -> str:
    """
    Returns the name of the profile

    That is, the path, relative to the profiles directory
    """
    return get_profile_path().split("profiles")[-1].lstrip(os.path.sep)


@prefix_aware_cache
def profile_parents() -> List[str]:
    """
    Produces the paths of all the parent directories for the selected profile, in order
    """
    if not env.PREFIX_NAME:
        return []

    repo, first = _profile_relative_path()

    def get_parents(base: str, directory: str, encoding=None) -> List[str]:
        parentfile = os.path.join(base, directory, "parent")
        parents = []
        if os.path.exists(parentfile):
            for parent in read_list(parentfile, encoding=encoding):
                parentpath = os.path.normpath(os.path.join(directory, parent))
                parents.extend(get_parents(base, parentpath))
                parents.append(parentpath)

        return parents

    parents = [first]
    parents.extend(get_parents(repo, first, encoding="utf-8"))

    userpath = os.path.join(env.prefix().CONFIG_DIR, "profile.user")
    if os.path.exists(userpath):
        parents.append(userpath)
        parents.extend(get_parents(userpath, "."))

    return [os.path.join(repo, path) for path in parents]


def get_system() -> Set[Atom]:
    """Calculates the system set using the user's currently selected profile"""
    system: Set[Atom] = set()
    for parent in profile_parents():
        packages = os.path.join(parent, "packages")
        if os.path.exists(packages):
            system |= {
                Atom(mod.lstrip("*"))
                for mod in read_list(packages)
                if mod.startswith("*")
            }

    return system



#!/usr/bin/env python

# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3


import os
import subprocess
import sys

from setuptools import Command, find_packages, setup
from setuptools_rust import Binding, RustExtension, Strip

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as file:
    long_description = file.read()


def get_manpages():
    pages = []
    path = "doc/_build/man"
    if os.path.exists(path):
        for directory in os.listdir(path):
            pages.append(
                (
                    os.path.join("share/man", directory),
                    [
                        os.path.join(path, directory, file)
                        for file in os.listdir(os.path.join(path, directory))
                    ],
                )
            )

    return pages


class BuildMan(Command):
    """Command to generate man pages"""

    description = "build man pages"
    user_options = []  # type: ignore

    def initialize_options(self):
        ...

    def finalize_options(self):
        ...

    def run(self):
        if sys.platform != "win32":
            # Don't build docs if sphinx isn't available
            try:
                import sphinx  # noqa
                import sphinxarg  # noqa

                subprocess.check_call(["make", "-C", "doc", "man"])
            except ImportError:
                pass


setup(
    name="portmod",
    author="Portmod Authors",
    author_email="incoming+portmod-portmod-9660349-issue-@incoming.gitlab.com",
    description="A CLI package manager for mods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://gitlab.com/portmod/portmod",
    download_url="https://gitlab.com/portmod/portmod/-/releases",
    packages=find_packages(exclude=["*.test", "*.test.*", "test.*", "test"]),
    rust_extensions=[
        RustExtension("portmodlib.portmod", binding=Binding.PyO3, strip=Strip.Debug)
    ],
    zip_safe=False,
    entry_points=(
        {
            "console_scripts": [
                "inquisitor = portmod._cli.inquisitor:main",
                "portmod= portmod._cli.main:main",
            ]
        }
    ),
    cmdclass={"build_man": BuildMan},
    data_files=get_manpages(),
    python_requires=">=3.7",
    install_requires=[
        "colorama",
        "GitPython",
        "progressbar2>=3.2",
        'pywin32; platform_system == "Windows"',
        "RestrictedPython>=4.0",
        "redbaron",
        # RC2+minicard is broken in these versions
        "python-sat>=0.1.5.dev0,!=0.1.7.dev22,!=0.1.7.dev23,!=0.1.7.dev24",
        'python-sat>=0.1.5.dev12; platform_system == "Windows"',
        "requests",
        'importlib_metadata; python_version < "3.8"',
        "packaging",
        "fasteners>=0.16",
        # Deprecated. Only needed by Pybuild1
        "chardet",
        "patool",
    ],
    setup_requires=["setuptools_scm", 'wheel; platform_system == "Windows"'],
    use_scm_version={"write_to": "portmod/_version.py"},
    extras_require={
        "dev": ["black", "flake8", "pylint", "isort", "mypy"],
        "test": ["pytest", "pytest-cov", "setuptools_scm"],
        "benchmark": ["pytest-benchmark"],
        # Bash completion support
        "bash": ["argcomplete"],
        # For building manpages
        "man": ["sphinx", "sphinx-argparse"],
        # For building html documentation
        "doc": [
            "sphinx",
            "sphinx-argparse",
            "sphinx-autodoc-typehints",
            "sphinx_rtd_theme",
            "sphinxcontrib-apidoc",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Games/Entertainment",
        "Topic :: System :: Software Distribution",
    ],
)
